"""Exec module for managing Amazon API Gateway Stage."""


async def get(
    hub,
    ctx,
    resource_id: str,
):
    """Get an API Gateway Stage from AWS. If more than one resource is found, the first resource returned from AWS
        will be used. The function returns None when no resource is found.

    Args:
        resource_id(str):
            Idem Resource ID that is generated once the resource is created,
                formatted as: <rest_api_id>-<name>.

    Returns:
        .. code-block:: python

            {"result": True|False, "comment": A message List, "ret": Dict[str, Any]}

    Examples:

        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.apigateway.stage.get resource_id="resource_id"


        Calling this exec module function from within a state module in pure python

        .. code-block:: yaml

            my_unmanaged_stack:
              exec.run:
                - path: aws.apigateway.stage.get
                - kwargs:
                    resource_id: '<rest_api_id>-<name>'

    """

    result = dict(comment=[], ret=None, result=True)
    if resource_id and len(resource_id.split("-")) == 2:
        rest_api_id, name = resource_id.split("-")
        ret = await hub.exec.boto3.client.apigateway.get_stage(
            ctx, restApiId=rest_api_id, stageName=name
        )

        if not ret["result"]:
            if "NotFoundException" in str(ret["comment"]):
                result["comment"].append(
                    hub.tool.aws.comment_utils.get_empty_comment(
                        resource_type="aws.apigateway.stage", name=name
                    )
                )
                result["comment"] += list(ret["comment"])
                return result
            result["comment"] += list(ret["comment"])
            result["result"] = False
            return result

        if not ret["ret"]:
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigateway.stage", name=name
                )
            )
            return result

        else:
            result["ret"] = hub.tool.aws.apigateway.stage.convert_raw_stage_to_present(
                raw_resource=ret["ret"],
                resource_id=resource_id,
            )

        return result

    else:
        result["comment"].append(
            f"Invalid Resource ID '{resource_id}'. Resource ID should be of format <rest_api_id>-<name>.",
        )
        result["result"] = False
        return result



"""
Exec module for rds.db_cluster
"""
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__func_alias__ = {"list_": "list"}


async def get(hub, ctx, resource_id: str, name: str = None) -> Dict[str, Any]:
    """
    Returns information about Amazon Aurora DB clusters and Multi-AZ DB clusters. For more information on Amazon Aurora
    DB clusters, see  What is Amazon Aurora? in the Amazon Aurora User Guide. For more information on Multi-AZ DB
    clusters, see  Multi-AZ deployments with two readable standby DB instances in the Amazon RDS User Guide. This
    operation can also return information for Amazon Neptune DB instances and Amazon DocumentDB instances.

        Args:
            resource_id (str):
                An identifier of the resource in the provider. Defaults to None.

            name(str, Optional): Idem name of the resource. Defaults to None.

        Examples:

            .. code-block:: bash

                idem exec aws_auto.rds.db_cluster.get.get resource_id=0123456789abcdef

            .. code-block:: python

                async def my_func(hub, ctx):
                    ret = await hub.exec.aws_auto.rds.db_cluster.get.get(ctx, resource_id=0123456789abcdef)

            .. code-block:: yaml

                aws_auto_rds_db_cluster_get_resource:
                  exec.run:
                    - path: aws_auto.rds.db_cluster.get
                    - kwargs:
                        resource_id: 0123456789abcdef

    """

    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    ret = await hub.exec.boto3.client.rds.describe_db_clusters(
        ctx=ctx,
        DBClusterIdentifier=resource_id,
    )

    if not ret["result"]:
        if "DBClusterNotFoundFault" in str(ret.get("comment", "")):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.rds.db_cluster",
                    name=name if name else resource_id,
                )
            )
        else:
            ret["result"] = False
        result["comment"] += list(ret["comment"])
        return result
    if not ret["ret"]["DBClusters"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_cluster",
                name=name if name else resource_id,
            )
        )
        return result

    resource = ret["ret"]["DBClusters"][0]
    if len(ret["ret"]["DBClusters"]) > 1:
        result["comment"].append(
            f"More than one aws.rds.db_cluster resource was found. Use resource {resource.get('DBClusterIdentifier')}"
        )
    result["ret"] = hub.tool.aws.rds.conversion_utils.convert_raw_db_cluster_to_present(
        idem_resource_name=name if name else resource_id,
        raw_resource=resource,
    )
    return result


async def list_(
    hub,
    ctx,
    filters: List[
        make_dataclass("Filter", [("Name", str), ("Values", List[str])])
    ] = None,
) -> Dict[str, Any]:
    """
    Returns information about Amazon Aurora DB clusters and Multi-AZ DB clusters. For more information on Amazon Aurora
    DB clusters, see  What is Amazon Aurora? in the Amazon Aurora User Guide. For more information on Multi-AZ DB
    clusters, see  Multi-AZ deployments with two readable standby DB instances in the Amazon RDS User Guide. This
    operation can also return information for Amazon Neptune DB instances and Amazon DocumentDB instances.

        Args:
            filters (list[Dict[str, Any]], Optional):
                A filter that specifies one or more DB clusters to describe.

                Supported filters:

                    * clone-group-id:
                        Accepts clone group identifiers. The results list only includes information about the DB
                        clusters associated with these clone groups.

                    * db-cluster-id:
                        Accepts DB cluster identifiers and DB cluster Amazon Resource Names (ARNs). The results list
                        only includes information about the DB clusters identified by these ARNs.

                    * domain:
                        Accepts Active Directory directory IDs.  The results list only includes information about the
                        DB clusters associated with these domains.

                    * engine:
                        Accepts engine names. The results list only includes information about the DB clusters for
                        these engines. Defaults to None.

                * Name (str):
                    The name of the filter. Filter names are case-sensitive.

                * Values (list[str]):
                    One or more filter values. Filter values are case-sensitive.

        Examples:

            .. code-block:: bash

                idem exec aws_auto.rds.db_cluster.list

            .. code-block:: python

                async def my_func(hub, ctx):
                    ret = await hub.exec.aws_auto.rds.db_cluster.list(ctx)

            .. code-block:: yaml

                aws_auto_rds_db_cluster_list_resource:
                  exec.run:
                    - path: aws_auto.rds.db_cluster.list

    """

    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    ret = await hub.exec.boto3.client.rds.describe_db_clusters(
        ctx=ctx,
        filters=filters,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        return result
    if not ret["ret"]["DBClusters"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_cluster",
                name="rds.db_cluster.list",
            )
        )
        return result

    result["ret"] = []
    for resource in ret["ret"]["DBClusters"]:
        result["ret"].append(
            hub.tool.aws.rds.conversion_utils.convert_raw_db_cluster_to_present(
                raw_resource=resource
            )
        )
    return result



"""Exec module for managing Rds Db Proxy Targets."""
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["soft_fail"]

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    db_proxy_name: str = None,
    resource_id: str = None,
    target_group_name: str = None,
    name: str = None,
) -> Dict[str, Any]:
    """

    Returns information about DBProxyTarget objects. This API supports pagination.

    Args:
        db_proxy_name(str, Optional): The identifier of the DBProxyTarget to describe.

        resource_id(str, Optional): Db_proxy_target unique ID.

        target_group_name(str, Optional): The identifier of the DBProxyTargetGroup to describe. Defaults to None.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            unmanaged_resource:
              exec.run:
                - path: aws.rds.db_proxy_target.get
                - kwargs:
                  db_proxy_name: value
                  resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_target.get db_proxy_name=value, resource_id=value
    """

    result = dict(comment=[], ret=None, result=True)
    if resource_id:
        resource_id_arr = resource_id.split("/")
        if len(resource_id_arr) < 4:
            result["result"] = False
            result["comment"].append(
                "Invalid resource_id. resource_id should be in format {db_proxy_name}/{target_group_name}/{target_group_type}/{rds_instance_id}"
            )
            return result
        db_proxy_name = resource_id_arr[0]
        target_group_name = resource_id_arr[1]
    db_proxy_target_get = await hub.exec.boto3.client.rds.describe_db_proxy_targets(
        ctx=ctx,
        **{"DBProxyName": db_proxy_name, "TargetGroupName": target_group_name},
    )

    # Case: Error
    if not db_proxy_target_get["result"]:
        # Do not return success=false when it is not found.
        # Most of the resources would return "*NotFound*" type of exception when it is 404
        if (
            "DBProxyNotFoundFault" in str(db_proxy_target_get["comment"])
            or "DBProxyTargetNotFoundFault" in str(db_proxy_target_get["comment"])
            or "DBProxyTargetGroupNotFoundFault" in str(db_proxy_target_get["comment"])
        ):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.rds.db_proxy_target", name=resource_id
                )
            )
            result["comment"].append(db_proxy_target_get["comment"])
            return result

        result["comment"].append(db_proxy_target_get["comment"])
        result["result"] = False
        return result

    # Case: Empty results
    if not db_proxy_target_get["ret"] or not db_proxy_target_get["ret"]["Targets"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_proxy_target", name=resource_id
            )
        )
        return result

    if len(db_proxy_target_get["ret"]["Targets"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_id=resource_id, resource_type="aws.rds.db_proxy_target"
            )
        )

    raw_resource = db_proxy_target_get["ret"]["Targets"][0]
    result[
        "ret"
    ] = await hub.tool.aws.rds.db_proxy_target.convert_raw_resource_to_present_async(
        ctx=ctx,
        resource_id=resource_id,
        raw_resource=raw_resource,
        idem_resource_name=name,
        db_proxy_name=db_proxy_name,
        target_group_name=target_group_name,
    )

    return result


async def list_(
    hub,
    ctx,
    db_proxy_name: str,
    target_group_name: str = "default",
    name: str = None,
) -> Dict[str, Any]:
    """
    Returns information about DBProxyTarget objects. This API supports pagination.

    Args:
        db_proxy_name(str): The identifier of the DBProxy to describe.

        target_group_name(str): The identifier of the DBProxyTargetGroup to describe. Defaults to None.

        filters(List[dict[str, Any]], Optional): This parameter is not currently supported. Defaults to None.

            * Name (str): The name of the filter. Filter names are case-sensitive.

            * Values (List[str]): One or more filter values. Filter values are case-sensitive.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        Resource State:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.rds.db_proxy_target.list
                - kwargs:
                  db_proxy_name: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_target.list db_proxy_name=value

        Describe call from the CLI:

        .. code-block:: bash

            $ idem describe aws.rds.db_proxy_target

    """

    result = dict(comment=[], ret=[], result=True)

    ret = await hub.exec.boto3.client.rds.describe_db_proxy_targets(
        ctx=ctx,
        **{
            "DBProxyName": db_proxy_name,
            "TargetGroupName": target_group_name,
        },
    )

    if not ret["result"]:
        result["comment"].append(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"].get("Targets"):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.rds.db_proxy_target", name=None
            )
        )
        return result

    for resource in ret["ret"]["Targets"]:
        resource_id = hub.tool.aws.rds.db_proxy_target.create_resource_id(
            resource, db_proxy_name, target_group_name
        )
        result["ret"].append(
            await hub.tool.aws.rds.db_proxy_target.convert_raw_resource_to_present_async(
                ctx=ctx,
                resource_id=resource_id,
                raw_resource=resource,
                idem_resource_name=name if name else resource_id,
                db_proxy_name=db_proxy_name,
                target_group_name=target_group_name,
            )
        )
    return result


async def create(
    hub,
    ctx,
    db_proxy_name: str,
    target_group_name: str,
    db_instance_identifiers: List[str] = None,
    db_cluster_identifiers: List[str] = None,
    resource_id: str = None,
    name: str = None,
) -> Dict[str, Any]:
    """
    Associate one or more DBProxyTarget data structures with a DBProxyTargetGroup.

    Args:
        db_proxy_name(str): The identifier of the DBProxy that is associated with the DBProxyTargetGroup.

        target_group_name(str, Optional): The identifier of the DBProxyTargetGroup. Defaults to None.

        db_instance_identifiers(List[str], Optional): One or more DB instance identifiers. Defaults to None.

        db_cluster_identifiers(List[str], Optional): One or more DB cluster identifiers. Defaults to None.

        resource_id(str, Optional): Db_proxy_target unique ID. Defaults to None.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Using in a state:

        .. code-block:: sls

            resource_is_present:
              aws.rds.db_proxy_target.present:
                - db_proxy_name: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_target.create db_proxy_name=value
    """

    result = dict(comment=[], ret={}, result=True)

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    resource_to_raw_input_mapping = {
        "db_proxy_name": "DBProxyName",
        "target_group_name": "TargetGroupName",
        "db_instance_identifiers": "DBInstanceIdentifiers",
        "db_cluster_identifiers": "DBClusterIdentifiers",
    }

    payload = {}
    for key, value in desired_state.items():
        if key in resource_to_raw_input_mapping.keys() and value is not None:
            payload[resource_to_raw_input_mapping[key]] = desired_state.get(key)

    ret = await hub.exec.boto3.client.rds.register_db_proxy_targets(ctx, **payload)

    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"].append(ret["comment"])
        return result

    result["comment"].append(
        f"Created aws.rds.db_proxy_target '{name}'",
    )

    raw_resource = ret["ret"].get("DBProxyTargets")[0]
    result["ret"]["resource_id"] = hub.tool.aws.rds.db_proxy_target.create_resource_id(
        raw_resource, db_proxy_name, target_group_name
    )
    result["ret"]["name"] = name

    return result


async def delete(
    hub,
    ctx,
    resource_id: str,
    db_instance_identifiers: List[str] = None,
    db_cluster_identifiers: List[str] = None,
    name: str = None,
) -> Dict[str, Any]:
    """
    Remove the association between one or more DBProxyTarget data structures and a DBProxyTargetGroup.

    Args:
        resource_id(str): Db_proxy_target unique ID.

        db_instance_identifiers(List[str], Optional): One or more DB instance identifiers. Defaults to None.

        db_cluster_identifiers(List[str], Optional): One or more DB cluster identifiers. Defaults to None.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            resource_is_absent:
              aws.rds.db_proxy_target.absent:
                - db_proxy_name: value
                - resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_target.delete db_proxy_name=value, resource_id=value
    """

    result = dict(comment=[], ret=None, result=True)
    resource_id_arr = resource_id.split("/")
    if len(resource_id_arr) < 4:
        result["result"] = False
        result["comment"].append(
            "Invalid resource_id. resource_id should be in format {db_proxy_name}/{target_group_name}/{target_group_type}/{rds_instance_id}"
        )
        return result
    db_proxy_name = resource_id_arr[0]
    target_group_name = resource_id_arr[1]

    delete_ret = await hub.exec.boto3.client.rds.deregister_db_proxy_targets(
        ctx,
        **{
            "DBProxyName": db_proxy_name,
            "TargetGroupName": target_group_name,
            "DBInstanceIdentifiers": db_instance_identifiers,
            "DBClusterIdentifiers": db_cluster_identifiers,
        },
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        result["comment"] = delete_ret["comment"]
        result["result"] = False
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.rds.db_proxy_target", name=name
    )

    return result



"""Exec module for managing Amazon API Gateway Integration Responses."""


async def get(
    hub,
    ctx,
    resource_id: str,
    name: str = None,
):
    """
    Get an API Gateway Integration Response from AWS.

    Args:
        resource_id(str):
            Idem Resource ID that is generated once the resource is created,
                formatted as: <rest_api_id>-<parent_resource_id>-<http_method>-<status_code>

        name(str, Optional):
            An Idem name of the API Gateway Integration.

    Returns:
        .. code-block:: python

            {"result": True|False, "comment": A message List, "ret": Dict[str, Any]}

    Examples:

        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.apigateway.integration_response.get resource_id="resource_id" name=name


        Calling this exec module function from within a state module in pure python

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, name):
                ret = await hub.exec.aws.apigateway.integration_response.get(ctx, resource_id=resource_id, name=name)

    """

    result = dict(comment=[], ret=None, result=True)
    if resource_id and len(resource_id.split("-")) == 4:
        rest_api_id, parent_resource_id, http_method, status_code = resource_id.split(
            "-"
        )

        get_integration_response = (
            await hub.exec.boto3.client.apigateway.get_integration_response(
                ctx,
                restApiId=rest_api_id,
                resourceId=parent_resource_id,
                httpMethod=http_method,
                statusCode=status_code,
            )
        )

        if not get_integration_response["result"]:
            if "NotFoundException" in str(get_integration_response["comment"]):
                result["comment"].append(
                    hub.tool.aws.comment_utils.get_empty_comment(
                        resource_type="aws.apigateway.integration_response", name=name
                    )
                )
                result["comment"] += list(get_integration_response["comment"])
                return result
            result["comment"] += list(get_integration_response["comment"])
            result["result"] = False
            return result

        if get_integration_response["ret"]:
            result[
                "ret"
            ] = hub.tool.aws.apigateway.integration_response.convert_raw_integration_response_to_present(
                idem_resource_name=name,
                raw_resource=get_integration_response["ret"],
                resource_id=resource_id,
            )

        return result

    else:
        result["comment"] = (
            f"Invalid Resource ID '{resource_id}'. Resource ID should be of format <rest_api_id>-<parent_resource_id>-<http_method>-<status_code>.",
        )
        result["result"] = False
        return result



"""Exec module for caller identity."""
from typing import Any
from typing import Dict


async def get(hub, ctx) -> Dict[str, Any]:
    """Returns details about the IAM user whose credentials are used to call the operation.

    Returns:
        Dict[str, Any]:
            Return the identity details such as UserId, Account, and ARN about the IAM user

    Examples:
        Call from the CLI:

        .. code-block:: bash

            $ idem exec aws.sts.caller_identity.get

        Call from code:

        .. code-block:: python

            async def my_func(hub, ctx):
                await hub.exec.aws.sts.caller_identity.get(ctx)

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.sts.caller_identity.get
    """
    ret = dict(result=True, ret={}, comment="")
    caller_identity = await hub.exec.boto3.client.sts.get_caller_identity(ctx)

    ret["result"] = caller_identity.result
    ret["comment"] = caller_identity.comment
    if caller_identity.result:
        ret["ret"]["UserId"] = caller_identity.ret["UserId"]
        ret["ret"]["Account"] = caller_identity.ret["Account"]
        ret["ret"]["Arn"] = caller_identity.ret["Arn"]
    return ret



"""Exec module for managing Amazon API Rest APIs."""


async def get(hub, ctx, name: str, resource_id: str):
    """Returns a rest api, including their root resource id.

    Args:
        name (str): idem name associated with this resource_id.

        resource_id (str): id of the specific rest_api to look up.

    Returns:
        dict[str, Any]:
            Returns details of a rest_api.

    Examples:

        .. code-block:: bash

            idem exec aws.apigateway.rest_api.get name=test_rest_api resource_id=xvwtynovjx --output json

                Call from code:

        .. code-block:: python

            async def my_func(hub, ctx, name:str, resource_id:str):
                await hub.exec.aws.apigateway.rest_api.get(ctx, name, resource_id, api_id)

        Normally, rest_api would be managed via aws.apigateway.rest_api.present, but using get in a state is possible:

        .. code-block:: yaml

            my_unmanaged_rest_api:
              exec.run:
                - path: aws.apigateway.rest_api.get
                - kwargs:
                    name: my_resource
                    resource_id: resource_id

    """
    result = dict(comment=[], ret=None, result=True)

    # find the rest_api
    partial_rest_api = await hub.exec.boto3.client.apigateway.get_rest_api(
        ctx, restApiId=resource_id
    )
    if not partial_rest_api["result"]:
        result["comment"] += tuple(partial_rest_api["comment"])
        result["result"] = False
        return result

    # find the id for the root resource ("/").
    resources = await hub.exec.boto3.client.apigateway.get_resources(
        ctx, restApiId=resource_id
    )
    if not resources["result"]:
        result["comment"] += (
            f"Unable to load root resource for rest_api '{partial_rest_api['ret']['name']}'. "
            f"This is likely a temporary issue.",
        )
        result["comment"] += tuple(resources["comment"])
        result["result"] = False
        return result
    root_resource_id = None
    for resource in resources["ret"]["items"]:
        if resource["path"] == "/":
            root_resource_id = resource["id"]
            break
    partial_rest_api["ret"]["rootResourceId"] = root_resource_id

    # convert to expected format
    result["ret"] = hub.tool.aws.apigateway.rest_api.convert_raw_rest_api_to_present(
        partial_rest_api["ret"], idem_resource_name=name
    )

    return result



"""Read operation's metadata and extract function definition"""
import botocore.client
import botocore.docs.docstring
from dict_tools.data import NamespaceDict


def parse(
    hub,
    client,
    aws_service_name: str,
    resource_name: str,
    func_name: str,
):
    if not func_name:
        return {
            "doc": "",
            "params": {},
            "hardcoded": {
                "aws_service_name": aws_service_name,
                "resource_name": resource_name,
                "boto3_documentation": f"https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/{aws_service_name}.html",
            },
        }

    function = getattr(client, func_name)
    doc: botocore.docs.docstring.ClientMethodDocstring = function.__doc__
    docstring = hub.tool.format.html.parse(doc._gen_kwargs["method_description"])

    function_doc = "\n".join(hub.tool.format.wrap.wrap(docstring, width=112))
    is_idempotent = "idempotent" in function_doc.lower()

    # Parse request params
    parameters = hub.pop_create.aws.function.resolve_request_params(
        aws_service_name, func_name, doc
    )
    hub.pop_create.aws.function.normalize_name_param(parameters)

    function_input_param_names = list(parameters.keys())
    function_resource_to_input_param_mappings = dict(
        map(
            lambda x: (hub.tool.format.case.snake(x), x),
            parameters.keys(),
        )
    )

    # Parse response
    (
        response_key,
        raw_resource_mappings,
    ) = hub.pop_create.aws.function.resolve_response_metadata(
        aws_service_name, func_name, doc
    )

    possible_response_keys = []
    if not response_key:
        possible_response_keys = list(raw_resource_mappings.keys())

    # This is helpful when resolving params
    boto3_documentation = f"https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/{aws_service_name}/client/{func_name}.html"

    # Build function definition
    ret = {
        "doc": f"{function_doc}",
        "params": parameters,
        "hardcoded": {
            "aws_service_name": aws_service_name,
            "resource_name": resource_name,
            "function": func_name,
            "boto3_function": f"hub.exec.boto3.client.{aws_service_name}.{func_name}",
            "boto3_documentation": boto3_documentation,
            "function_input_param_names": function_input_param_names,
            "function_resource_to_input_param_mappings": function_resource_to_input_param_mappings,
            "response_key": response_key,
            "possible_response_keys": possible_response_keys,
            "resource_to_raw_resource_mappings": {
                v: k for k, v in raw_resource_mappings.items()
            },
            "raw_resource_to_resource_mapping": raw_resource_mappings,
            "resource_attributes": list(raw_resource_mappings.keys()),
            "has_client_token": bool(parameters.get("ClientToken", None)),
            "is_idempotent": is_idempotent,
        },
    }

    return ret


def resolve_request_params(
    hub,
    aws_service_name,
    func_name,
    doc: "botocore.docs.docstring.ClientMethodDocstring",
):
    parameters = NamespaceDict()
    try:
        input_param_struct = doc._gen_kwargs["operation_model"].input_shape

        if not input_param_struct:
            return parameters

        params = input_param_struct.members
        required_params = doc._gen_kwargs[
            "operation_model"
        ].input_shape.required_members

        # these params are not used in idem-aws, should be skipped
        unused_params = ["DryRun", "Marker", "MaxRecords", "MaxResults", "NextToken"]

        for p, data in params.items():
            # Unwrap TagSpecification and get tags
            if "TagSpecifications" in p:
                p = "tags"
                data = data.member.members["Tags"]
            elif p in unused_params:
                # skip these params as idem-aws don't make use of it
                continue
            parameters[p] = hub.pop_create.aws.param.parse(
                param=data, required=p in required_params, parsed_nested_params=[]
            )
    except AttributeError as e:
        hub.log.error(
            f"Error reading parameters for {aws_service_name}.{func_name}: {e}"
        )

    return parameters


def normalize_name_param(hub, parameters: dict):
    # Normalize the name parameter
    name = None
    if "Name" in parameters:
        name = parameters.pop("Name")
    elif "name" in parameters:
        name = parameters.pop("name")

    if bool(name):
        parameters["Name"] = name


def resolve_response_metadata(
    hub,
    aws_service_name,
    func_name,
    doc: "botocore.docs.docstring.ClientMethodDocstring",
):
    response_metadata = doc._gen_kwargs["operation_model"].output_shape

    if response_metadata is None:
        return None, {}

    response_key = None
    raw_resource_mappings = {}

    try:
        # these params are not used in idem-aws, should be skipped
        paginated_response_unused_params = [
            "Marker",
            "MaxRecords",
            "MaxResults",
            "NextToken",
            "TotalCount",
        ]

        response_params = response_metadata.members
        is_paginated_response = any(
            p in paginated_response_unused_params for p in response_params.keys()
        )
        if is_paginated_response:
            for p, data in response_params.items():
                if p in paginated_response_unused_params:
                    # skip "Marker" as idem-aws don't make use of it
                    continue

                if isinstance(data, botocore.model.StructureShape):
                    # Example:
                    #   {
                    #     'Certificate': {
                    #           .....
                    #      }
                    #   }
                    raw_resource_mappings = dict(
                        map(
                            lambda x: (x, hub.tool.format.case.snake(x)),
                            data.members.keys(),
                        )
                    )
                elif isinstance(data, botocore.model.ListShape):
                    # Example:
                    #   {
                    #     'Vpcs': [
                    #         {
                    #               .....
                    #         },
                    #     ]
                    #   }
                    raw_resource_mappings = dict(
                        map(
                            lambda x: (x, hub.tool.format.case.snake(x)),
                            data.member.members.keys(),
                        )
                    )

                # This should be the only remaining response param after discarding pagination params
                response_key = p
        else:
            raw_resource_mappings = dict(
                map(
                    lambda x: (x, hub.tool.format.case.snake(x)), response_params.keys()
                )
            )
    except AttributeError as e:
        hub.log.error(
            f"Error reading return fields for {aws_service_name}.{func_name}: {e}"
        )

    return response_key, raw_resource_mappings



import copy

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    parent_resource_id: str,
    http_method: str,
    input_type: str,
    resource_id: str = None,
    cache_namespace: str = None,
    cache_key_parameters: list = None,
    integration_responses: dict = None,
    timeout_in_millis: int = None,
    passthrough_behavior: str = None,
    connection_id: str = None,
    integration_http_method: str = None,
    uri: str = None,
):
    """
    Creates a new API Gateway Integration or modifies an existing one.

    Args:
        name(str):
            An idem name of the resource.

        rest_api_id(str):
            AWS rest_api id of the associated RestApi.

        parent_resource_id(str):
            The parent resource's id.

        http_method(str):
            String that specifies the method request's HTTP method type.

        input_type(str):
            Specifies a put integration input's type.

        resource_id(str, Optional):
            Idem resource id, formatted as: rest_api_id-parent_resource_id-http_method. Defaults to None.

        cache_namespace(str, Optional):
            Specifies a group of related cached parameters. By default, API Gateway uses
            the resource ID as the cacheNamespace . You can specify the same cacheNamespace across resources to return
            the same cached data for requests to different resources.

        cache_key_parameters(list, Optional):
            A list of request parameters whose values API Gateway caches. To be valid
            values for cacheKeyParameters , these parameters must also be specified for Method requestParameters.

        timeout_in_millis(int, Optional):
            Custom timeout between 50 and 29,000 milliseconds. The default value is 29,000 milliseconds or 29 seconds.

        passthrough_behavior:
            Custom timeout between 50 and 29,000 milliseconds. The default value is
            29,000 milliseconds or 29 seconds.

        connection_id(str, Optional):
            The ID of the VpcLink used for the integration. Specify this value only if you specify VPC_LINK as the
            connection type.

        integration_responses(dict, Optional):
            Specifies the integration's responses. The status code must map to an existing MethodResponse,
            and parameters and templates can be used to transform the back-end response.

        integration_http_method(string, Optional):
            The HTTP method for the integration.

        uri(string, Optional):
            Specifies Uniform Resource Identifier (URI) of the integration endpoint. For HTTP or HTTP_PROXY integrations, the URI must be a fully formed,
            encoded HTTP(S) URL according to the RFC-3986 specification, for either standard integration, where connectionType is not VPC_LINK ,
            or private integration, where connectionType is VPC_LINK . For a private HTTP integration, the URI is not used for routing.
            For AWS or AWS_PROXY integrations, the URI is of the form arn:aws:apigateway:{region}:{subdomain.service|service}:path|action/{service_api }.
            Here, {Region} is the API Gateway region (e.g., us-east-1); {service} is the name of the integrated Amazon Web Services service (e.g., s3);
            and {subdomain} is a designated subdomain supported by certain Amazon Web Services service for fast host-name lookup. action can be used for
            an Amazon Web Services service action-based API, using an Action={name}&{p1}={v1}&p2={v2}... query string.
            The ensuing {service_api} refers to a supported action {name} plus any required input parameters. Alternatively, path can be used for an
            Amazon Web Services service path-based API. The ensuing service_api refers to the path to an Amazon Web Services service resource,
            including the region of the integrated Amazon Web Services service, if applicable. For example, for integration with the S3 API of GetObject ,
            the uri can be either arn:aws:apigateway:us-west-2:s3:action/GetObject&Bucket={bucket}&Key={key} or arn:aws:apigateway:us-west-2:s3:path/{bucket}/{key} .

    Request Syntax:
        [idem_test_aws_apigateway_integration]:
          aws.apigateway.integration.present:
            - name: 'string'
            - rest_api_id: 'string'
            - parent_resource_id: 'string'
            - http_method: 'string'
            - input_type: 'string'

    Returns:
        dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_apigateway_integration:
              aws.apigateway.integration.present:
                - name: test_integration
                - rest_api_id: ${aws.apigateway.rest_api:idem_test_aws_apigateway_rest_api:resource_id}
                - parent_resource_id: ${aws.apigateway.rest_api:test_rest_api:root_resource_id}
                - http_method: ${aws.apigateway.method:test_method:http_method}
                - input_type: MOCK

            idem_test_aws_apigateway_method:
              aws.apigateway.method.present:
                - name: test_method
                - rest_api_id: ${aws.apigateway.rest_api:idem_test_aws_apigateway_rest_api:resource_id}
                - parent_resource_id: ${aws.apigateway.rest_api:idem_test_aws_apigateway_rest_api:root_resource_id}
                - http_method: GET
                - authorization_type: AWS_IAM

            idem_test_aws_apigateway_rest_api:
              aws.apigateway.rest_api.present:
                - name: test_rest_api

        Above, note that the integration's http_method is ``${aws.apigateway.method:test_method:http_method}``.
        This evaluates to ``GET``, plus an implicit ``require`` of the resource. You could also set it to ``GET``
        explicitly, and use an explicit ``require``:

        .. code-block:: sls

            idem_test_aws_apigateway_integration:
              aws.apigateway.integration.present:
                - require:
                    - aws.apigateway.method: test_method
                - http_method: GET
                ...

        .. code-block:: sls

            [idem_test_aws_apigateway_integration]:
              aws.apigateway.integration.present:
                - name: 'string'
                - rest_api_id: 'string'
                - parent_resource_id: 'string'
                - http_method: 'string'
                - input_type: 'string'

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.apigateway.integration.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

        result["old_state"] = before["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])

        update_parameters = {
            "cache_namespace": cache_namespace,
            "cache_key_parameters": cache_key_parameters,
            "integration_responses": integration_responses,
            "input_type": input_type,
            "timeout_in_millis": timeout_in_millis,
            "passthrough_behavior": passthrough_behavior,
            "connection_id": connection_id,
            "uri": uri,
        }

        update_ret = await hub.tool.aws.apigateway.integration.update(
            ctx,
            old_state=result["old_state"],
            updatable_parameters=update_parameters,
        )

        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        result["comment"] = result["comment"] + update_ret["comment"]
        resource_updated = bool(update_ret["ret"])

        if resource_updated and ctx.get("test", False):
            result["new_state"].update(update_ret["ret"])

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.apigateway.integration", name=name
                )
                return result

    else:
        resource_id = f"{rest_api_id}-{parent_resource_id}-{http_method}"

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "rest_api_id": rest_api_id,
                    "resource_id": resource_id,
                    "parent_resource_id": parent_resource_id,
                    "http_method": http_method,
                    "passthrough_behavior": passthrough_behavior,
                    "timeout_in_millis": timeout_in_millis,
                    "cache_namespace": cache_namespace,
                    "cache_key_parameters": cache_key_parameters,
                    "integration_responses": integration_responses,
                    "connection_id": connection_id,
                    "input_type": input_type,
                    "integration_http_method": integration_http_method,
                    "uri": uri,
                },
            )

            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.integration", name=name
            )
            return result

        ret = await hub.exec.boto3.client.apigateway.put_integration(
            ctx,
            restApiId=rest_api_id,
            resourceId=parent_resource_id,
            httpMethod=http_method,
            type=input_type,
            cacheNamespace=cache_namespace,
            cacheKeyParameters=cache_key_parameters,
            timeoutInMillis=timeout_in_millis,
            passthroughBehavior=passthrough_behavior,
            connectionId=connection_id,
            integrationHttpMethod=integration_http_method,
            uri=uri,
        )

        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.integration", name=name
        )
    if (not before) or resource_updated:
        after = await hub.exec.aws.apigateway.integration.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result

        resource_translated = after["ret"]
        result["new_state"] = resource_translated

    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
):
    """
    Deletes an API Gateway Integration.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            Idem resource id, formatted as: rest_api_id-parent_resource_id-http_method. Defaults to None.
            Idem automatically considers this resource being absent if this field is not specified.

    Examples:

        .. code-block:: sls

            idem_test_aws_apigateway_integration:
              aws.apigateway.integration.absent:
                - name: value
                - resource_id: value

    Returns:
        dict[str, Any]
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.integration", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigateway.integration.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.integration", name=name
        )
        return result

    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.integration", name=name
        )
        return result

    else:
        rest_api_id, parent_resource_id, http_method = resource_id.split("-")
        result["old_state"] = before_ret["ret"]
        delete_ret = await hub.exec.boto3.client.apigateway.delete_integration(
            ctx,
            restApiId=rest_api_id,
            resourceId=parent_resource_id,
            httpMethod=http_method,
        )

        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigateway.integration", name=name
        )
        return result


async def describe(hub, ctx):
    """
    Describe the API Gateway Integrations.

    Returns a list of apigateway.integration descriptions

    Returns:
        dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.apigateway.integration
    """
    result = {}

    get_rest_apis_ret = await hub.exec.boto3.client.apigateway.get_rest_apis(ctx)
    if not get_rest_apis_ret["result"]:
        hub.log.debug(f"Could not get Rest Apis {get_rest_apis_ret['comment']}")
        return result

    for rest_api in get_rest_apis_ret["ret"]["items"]:
        rest_api_id = rest_api.get("id")
        get_resources_ret = await hub.exec.boto3.client.apigateway.get_resources(
            ctx, restApiId=rest_api_id
        )
        if not get_resources_ret["result"]:
            f"Could not get Resources {get_resources_ret['comment']}. Will skip this Resource."
            continue

        if get_resources_ret["ret"]["items"] is not None:
            for resource in get_resources_ret["ret"]["items"]:
                parent_resource_id = resource.get("id")
                if resource.get("resourceMethods") is not None:
                    for resource_method in resource.get("resourceMethods"):
                        method_resource_id = (
                            f"{rest_api_id}-{parent_resource_id}-{resource_method}"
                        )

                        integration = await hub.exec.aws.apigateway.integration.get(
                            ctx,
                            resource_id=method_resource_id,
                        )

                        if not integration["ret"]:
                            hub.log.debug(
                                f"Could not get Integration {integration['comment']}. Will skip this Integration."
                            )
                            continue

                        resource_translated = integration["ret"]
                        result[resource_translated["resource_id"]] = {
                            "aws.apigateway.integration.present": [
                                {parameter_key: parameter_value}
                                for parameter_key, parameter_value in resource_translated.items()
                            ]
                        }

    return result



"""Build plugin metadata which can be used by pop-create for plugin code generation"""
from typing import Any
from typing import Dict

from dict_tools.data import NamespaceDict


def create_resource_plugin(
    hub,
    ctx,
    shared_resource_data: dict,
) -> Dict[str, Any]:
    """
    Create CloudSpecPlugin with exec/state/tool functions
    """
    aws_service_name = shared_resource_data.get("aws_service_name")

    plugin = {
        "doc": f"\nTODO: Update generated documentation based on official AWS docs for the service https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/{aws_service_name}.html",
        "imports": [
            "import copy",
            "from dataclasses import field",
            "from dataclasses import make_dataclass",
            "from dict_tools import differ",
            "from typing import List",
            "from typing import Any",
            "from typing import Dict",
        ],
        "functions": NamespaceDict(),
    }

    plugin["functions"]["get"] = hub.pop_create.aws.plugin.generate_get(
        shared_resource_data["functions"]
    )

    create = hub.pop_create.aws.plugin.generate_present(
        shared_resource_data["functions"]
    )
    plugin["functions"]["create"] = create
    plugin["functions"]["present"] = create

    list_ = hub.pop_create.aws.plugin.generate_list(shared_resource_data["functions"])
    plugin["functions"]["list"] = list_
    plugin["functions"]["describe"] = list_

    delete = hub.pop_create.aws.plugin.generate_absent(
        shared_resource_data["functions"]
    )
    plugin["functions"]["delete"] = delete
    plugin["functions"]["absent"] = delete

    plugin["functions"]["update"] = hub.pop_create.aws.plugin.generate_update(
        shared_resource_data["functions"]
    )

    plugin["functions"][
        "convert_raw_resource_to_present_async"
    ] = hub.pop_create.aws.plugin.generate_convert_raw_to_present(
        shared_resource_data.get("resource_name"),
        shared_resource_data["functions"].get("get"),
    )

    # Add all tools function
    non_reserved_functions = {
        func_name: func_data
        for func_name, func_data in shared_resource_data["functions"].items()
        if func_name not in ["get", "create", "delete", "update", "list"]
    }

    for func_name, func_data in non_reserved_functions.items():
        plugin["functions"][func_name] = func_data

    # Add all tests functions
    plugin = hub.pop_create.aws.plugin.add_tests_functions(plugin)

    return plugin


def create_tags_plugin(
    hub,
    aws_service_name: str,
    tag_functions: dict,
) -> Dict[str, Any]:
    """
    Create tags CloudSpecPlugin for the service
    """
    plugin = {
        "doc": f"Tags related functions for AWS service '{aws_service_name}'.",
        "imports": [
            "import copy",
            "from dataclasses import field",
            "from dataclasses import make_dataclass",
            "from typing import List",
            "from typing import Any",
            "from typing import Dict",
        ],
        "functions": NamespaceDict(
            get_tags_for_resource=hub.pop_create.aws.plugin.generate_get_tags_for_resource(
                tag_functions
            ),
            update_tags=hub.pop_create.aws.plugin.generate_update_tags(tag_functions),
        ),
    }

    return plugin


def create_service_plugin(
    hub,
    aws_service_name: str,
    service_functions: dict,
) -> Dict[str, Any]:
    """
    Create CloudSpecPlugin for the service
    """
    plugin = {
        "doc": f"Functions for AWS service '{aws_service_name}'.",
        "imports": [
            "from dataclasses import field",
            "from dataclasses import make_dataclass",
            "from typing import List",
            "from typing import Any",
            "from typing import Dict",
        ],
        "functions": service_functions,
    }

    return plugin


def generate_get(hub, shared_resource_data):
    get_function_definition = shared_resource_data.get("get", {})
    params = get_function_definition.get("params", {})
    return {
        "doc": f"\n{get_function_definition.get('doc', '')}",
        "params": params,
        "hardcoded": dict(
            **get_function_definition.get("hardcoded", {}),
        ),
    }


def generate_list(hub, shared_resource_data):
    describe_function_definition = shared_resource_data.get("list", {})
    params = describe_function_definition.get("params", {})
    return {
        "doc": f"{describe_function_definition.get('doc', '')}",
        "params": params,
        "hardcoded": dict(
            **describe_function_definition.get("hardcoded", {}),
        ),
    }


def generate_update(hub, shared_resource_data):
    update_function_definition = shared_resource_data.get("update", {})
    params = update_function_definition.get("params", {})
    if "Name" in params:
        # it gets added by default in header params by pop-create
        params.pop("Name")
    if "tags" not in params.keys():
        params["tags"] = hub.pop_create.aws.known_params.TAGS_PARAMETER.copy()
    return {
        "doc": f"{update_function_definition.get('doc', '')}",
        "params": params,
        "hardcoded": dict(
            **update_function_definition.get("hardcoded", {}),
        ),
    }


def generate_present(hub, shared_resource_data):
    create_function_definition = shared_resource_data.get("create", {})
    params = create_function_definition.get("params", {})
    if "Name" in params:
        # it gets added by default in header params by pop-create
        params.pop("Name")

    if "Tags" not in params.keys():
        params["tags"] = hub.pop_create.aws.known_params.TAGS_PARAMETER.copy()
    return {
        "doc": f"{create_function_definition.get('doc', '')}",
        "params": params,
        "hardcoded": dict(
            **create_function_definition.get("hardcoded", {}),
        ),
    }


def generate_absent(hub, shared_resource_data):
    delete_function_definition = shared_resource_data.get("delete", {})
    return {
        "doc": f"{delete_function_definition.get('doc', '')}",
        "params": delete_function_definition.get("params", {}),
        "hardcoded": dict(
            **delete_function_definition.get("hardcoded", {}),
        ),
    }


def generate_get_tags_for_resource(hub, tag_functions):
    list_tags_function_definition = tag_functions.get("list_tags", {})
    return {
        "doc": f"Get tags for a given resource.\n",
        "params": dict(
            resource_id=hub.pop_create.aws.known_params.RESOURCE_ID_PARAMETER.copy(),
        ),
        "hardcoded": dict(
            **list_tags_function_definition.get("hardcoded", {}),
        ),
    }


def generate_update_tags(hub, tag_functions):
    update_tags_function_definition = tag_functions.get("update_tags", {})
    if update_tags_function_definition:
        return {
            "doc": f"Updates tags for a given resource.\n",
            "params": update_tags_function_definition["params"],
            "hardcoded": dict(
                update_tags_boto3_function=update_tags_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_function"),
                update_tags_input_params=update_tags_function_definition.get(
                    "params", {}
                ).keys(),
                single_update=True,
            ),
        }
    else:
        untag_resource_function_definition = tag_functions.get("untag_resource", {})
        tag_resource_function_definition = tag_functions.get("tag_resource", {})
        return {
            "doc": f"Updates tags for a given resource.\n",
            "params": dict(
                resource_id=hub.pop_create.aws.known_params.RESOURCE_ID_PARAMETER.copy(),
                old_tags=hub.pop_create.aws.known_params.OLD_TAGS_PARAMETER.copy(),
                new_tags=hub.pop_create.aws.known_params.NEW_TAGS_PARAMETER.copy(),
            ),
            "hardcoded": dict(
                untag_resource_boto3_function=untag_resource_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_function"),
                untag_resource_input_params=list(
                    untag_resource_function_definition.get("params", {}).keys()
                ),
                untag_resource_documentation=untag_resource_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_documentation"),
                tag_resource_boto3_function=tag_resource_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_function"),
                tag_resource_input_params=list(
                    tag_resource_function_definition.get("params", {}).keys()
                ),
                tag_resource_documentation=tag_resource_function_definition.get(
                    "hardcoded", {}
                ).get("boto3_documentation"),
                single_update=False,
            ),
        }


def generate_convert_raw_to_present(
    hub, resource_name: str, get_function_definition: dict
):
    return {
        "doc": f"Convert raw resource of {resource_name} type into present format.\n",
        "params": dict(
            idem_resource_name=hub.pop_create.aws.known_params.NAME_PARAMETER.copy(),
            resource_id=hub.pop_create.aws.known_params.RESOURCE_ID_PARAMETER.copy(),
            raw_resource=hub.pop_create.aws.known_params.RAW_RESOURCE_PARAMETER.copy(),
        ),
        "hardcoded": dict(
            **get_function_definition.get("hardcoded", {}),
        )
        if get_function_definition
        else {},
    }


def add_tests_functions(hub, plugin: dict):
    functions = plugin.get("functions", {}).copy()
    for func_name, func_data in functions.items():
        if func_name in ["get", "list", "create", "update", "delete"]:
            test_module_type = "exec"
        elif func_name in ["present", "absent", "describe"]:
            test_module_type = "states"
        else:
            test_module_type = "tool"

        plugin["functions"][f"test_{func_name}"] = {
            "doc": "",
            "hardcoded": {
                "resource_name": func_data.get("hardcoded", {}).get("resource_name"),
                "aws_service_name": func_data.get("hardcoded", {}).get(
                    "aws_service_name"
                ),
                # This is critical to derive rendering location
                "test_module_type": test_module_type,
                # Add calling method's metadata such as name, call params etc.
                "method_call_name": func_name,
                "method_call_params": func_data.get("params"),
            },
        }

    return plugin



"""State module for managing AWS Api Gateway Base Path Mapping resources."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    domain_name: str,
    base_path: str = None,
    stage: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Creates a new resource or modifies an existing one.

    Args:
        name(str): An idem name of the resource.
        rest_api_id(str): AWS rest api id of the associated Api.
        domain_name(str): The domain name of the BasePathMapping resource to create.
        base_path(str, Optional): The base path name that callers of the API must provide as part of the URL after the domain name.
                            This value must be unique for all of the mappings across a single API.
                            Specify '(none)' if you do not want callers to specify a base path name after the domain name.
        stage(str, Optional): The name of the API's stage that you want to use for this mapping.
                                Specify '(none)' if you want callers to explicitly specify the stage name after any base path name.
        resource_id(str, Optional): Idem Resource ID that is generated once the resource is created..

    Request Syntax:
        [idem_test_aws_apigateway_resource]:
          aws.apigateway.resource.present:
            - name: 'string'
            - rest_api_id: 'string'
            - domain_name: 'string'
            - stage: 'string'
            - base_path: 'string'

    Returns:
        Dict[str, Any]

    Example:
        [idem_test_aws_apigateway_resource]:
          aws.apigateway.resource.present:
            - name: api.idem-jedi-test.com
            - rest_api_id: 1s6c0ucab9
            - domain_name: api.idem-jedi-test.com
            - stage: staging
            - base_path: idem
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    before_ret = None
    if resource_id:
        before_ret = await hub.exec.aws.apigateway.base_path_mapping.get(
            ctx,
            domain_name=domain_name,
            base_path=base_path,
            resource_id=resource_id,
        )
        if not before_ret["result"] or not before_ret["ret"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.base_path_mapping", name=domain_name
        )
        result["old_state"] = before_ret["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])
        update_parameters = {
            "rest_api_id": rest_api_id,
            "stage": stage,
            "base_path": base_path,
        }
        update_base_path = await hub.tool.aws.apigateway.base_path_mapping.update(
            ctx,
            domain_name=domain_name,
            old_state=result["old_state"],
            updatable_parameters=update_parameters,
        )
        if not update_base_path["result"]:
            result["result"] = False
            result["comment"] = update_base_path["comment"]
            return result
        result["comment"] = result["comment"] + update_base_path["comment"]
        resource_updated = bool(update_base_path["ret"])
        if resource_updated and ctx.get("test", False):
            result["new_state"].update(update_base_path["ret"])
        if resource_updated:
            if ctx.get("test", False):
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.apigateway.base_path_mapping", name=name
                )
                return result
            else:
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.apigateway.base_path_mapping", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "domain_name": domain_name,
                    "rest_api_id": rest_api_id,
                    "stage": stage,
                    "base_path": base_path,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.base_path_mapping", name=name
            )
            return result

        ret = await hub.exec.boto3.client.apigateway.create_base_path_mapping(
            ctx,
            restApiId=rest_api_id,
            domainName=domain_name,
            stage=stage,
            basePath=base_path,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.base_path_mapping", name=domain_name
        )
    if (not before_ret) or resource_updated:
        resource_id = f"{domain_name}/{base_path}"
        get_ret = await hub.exec.aws.apigateway.base_path_mapping.get(
            ctx,
            resource_id=resource_id,
        )
        if not (get_ret["result"] and get_ret["ret"]):
            result["result"] = False
            result["comment"] = get_ret["comment"]
            return result
        result["new_state"] = get_ret["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    domain_name: str = None,
    base_path: str = None,
    resource_id: str = None,
):
    """
    Deletes the specified base path mapping.

    Args:
        name(str): The Idem name of the resource.
        domain_name(str, Optional): the domain name of the BasePathMapping resource to delete.
        base_path(str, Optional): The base path name of the BasePathMapping resource to delete.
                                    To specify an empty base path, set this parameter to '(none)' .
        resource_id(str, Optional): AWS Resource id. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigateway_base_path_mapping:
              aws.apigateway.base_path_mapping.absent:
                - name: value
                - resource_id: value


    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.base_path_mapping", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigateway.base_path_mapping.get(
        ctx, resource_id=resource_id
    )

    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.base_path_mapping", name=name
        )
        return result

    if not domain_name or not base_path:
        result["comment"] = hub.tool.aws.comment_utils.missing_args_for_absent_comment(
            resource_type="aws.apigateway.base_path_mapping",
            name=name,
            args=["domain_name", "base_type"],
        )
        result["result"] = False
        return result

    if ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.base_path_mapping", name=name
        )
        return result

    else:
        result["old_state"] = before_ret["ret"]
        delete_ret = await hub.exec.boto3.client.apigateway.delete_base_path_mapping(
            ctx,
            domainName=domain_name,
            basePath=base_path,
        )

        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigateway.base_path_mapping", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the API Gateway base path mapping associated with a specific API.

    Returns a list of apigateway.base_path_mapping descriptions

    Returns:
        Dict[str, Any]


    Examples:

        .. code-block:: bash

            $ idem describe aws.apigateway.base_path_mapping

    """

    result = {}

    get_domain_names_ret = await hub.exec.boto3.client.apigateway.get_domain_names(ctx)
    if not get_domain_names_ret["result"]:
        hub.log.debug(f"Could not get domain_names {get_domain_names_ret['comment']}")
        return result

    for domain_name in get_domain_names_ret["ret"]["items"]:
        domain_name = domain_name.get("domainName")
        get_base_path_mappings_ret = (
            await hub.exec.aws.apigateway.base_path_mapping.list(
                ctx,
                domain_name=domain_name,
            )
        )
        if not get_base_path_mappings_ret["result"]:
            hub.log.debug(
                f"Could not get base path mappings for domain name '{domain_name}': "
                f"{get_base_path_mappings_ret['comment']}. Describe will skip this domain name and continue."
            )
            continue

        for base_path_mapping in get_base_path_mappings_ret["ret"]:
            result[base_path_mapping["resource_id"]] = {
                "aws.apigateway.base_path_mapping.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in base_path_mapping.items()
                ]
            }
    return result



from typing import List


async def apply(
    hub, ctx, resource, *, old_value, new_value, comments: List[str]
) -> bool:
    """
    This attribute is modified by the meta "init" function
    """
    return True



"""
State module for managing EC2 VPc Endpoints.
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    vpc_id: str,
    service_name: str,
    resource_id: str = None,
    vpc_endpoint_type: str = None,
    policy_document: str = None,
    route_table_ids: List[str] = None,
    subnet_ids: List[str] = None,
    security_group_ids: List[str] = None,
    private_dns_enabled: bool = None,
    client_token: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates a VPC endpoint for a specified service.

    An endpoint enables you to create a private connection between your VPC and the service. The service may be provided
    by Amazon Web Services, an Amazon Web Services Marketplace Partner, or another Amazon Web Services account. For more
    information, see VPC Endpoints in the Amazon Virtual Private Cloud User Guide. A gateway endpoint serves as a target
    for a route in your route table for traffic destined for the Amazon Web Service. You can specify an endpoint policy
    to attach to the endpoint, which will control access to the service from your VPC. You can also specify the VPC route
    tables that use the endpoint. An interface endpoint is a network interface in your subnet that serves as an endpoint
    for communicating with the specified service. You can specify the subnets in which to create an endpoint, and the
    security groups to associate with the endpoint network interface. A GatewayLoadBalancer endpoint is a network
    interface in your subnet that serves an endpoint for communicating with a Gateway Load Balancer that you've configured
    as a VPC endpoint service. Use DescribeVpcEndpointServices to get a list of supported services.

    Args:
        name(str):
            An Idem name of the resource.

        vpc_id(str):
            AWS VPC ID.

        resource_id(str, Optional):
            AWS VPC Endpoint id.

        service_name(str): The service name.
            To get a list of available services, use the DescribeVpcEndpointServices request, or get the name from the
            service provider.

        vpc_endpoint_type(str, Optional):
            The type of endpoint. Default: Gateway. Defaults to Gateway.

        policy_document(str, Optional):
            (Interface and gateway endpoints) A policy to attach to the endpoint that controls access to the service.
            The policy must be in valid JSON format. If this parameter is not specified, we attach a default policy that
            allows full access to the service. Defaults to None.

        route_table_ids(list, Optional):
            (Gateway endpoint) One or more route table IDs. Defaults to None.

        subnet_ids(list, Optional):
            (Interface and Gateway Load Balancer endpoints) The ID of one or more subnets in which to create an endpoint
            network interface. For a Gateway Load Balancer endpoint, you can specify one subnet only. Defaults to None.

        security_group_ids(list, Optional):
            (Interface endpoint) The ID of one or more security groups to associate with the endpoint
            network interface. Defaults to None.

        private_dns_enabled(bool, Optional):
            (Interface endpoint) Indicates whether to associate a private hosted zone with the specified VPC. The private
            hosted zone contains a record set for the default public DNS name for the service for the Region (for example,
            kinesis.us-east-1.amazonaws.com), which resolves to the private IP addresses of the endpoint network
            interfaces in the VPC. This enables you to make requests to the default public DNS name for the service instead
            of the public DNS names that are automatically generated by the VPC endpoint service. To use a private hosted
            zone, you must set the following VPC attributes to true: enableDnsHostnames and enableDnsSupport. Use
            ModifyVpcAttribute to set the VPC attributes. Default: true. Defaults to None.

        client_token(str, Optional):
            Unique, case-sensitive identifier that you provide to ensure the idempotency of the request. Constraint: Maximum 64 ASCII characters. This field is autopopulated if not provided.

        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of ``[{"Key": tag-key, "Value": tag-value}]``
            to associate with the VPC endpoint. Each tag consists of a key name and an associated value. Defaults to None.

            * Key (str, Optional):
                The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode characters.
                May not begin with aws:.

            * Value(str, Optional):
                The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

        timeout(dict, Optional):
            Timeout configuration for create/update of AWS VPC Endpoint.

            * create (dict):
                Timeout configuration for creating AWS VPC Endpoint

                * delay (int, Optional):
                    The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional):
                    Customized timeout configuration containing delay and max attempts.

            * update(dict, Optional):
                Timeout configuration for updating AWS VPC Endpoint

                * delay (int, Optional):
                    The amount of time in seconds to wait between attempts.

                * max_attempts (int, Optional):
                    Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        .. code-block:: sls

            [vpc_endpoint-id]:
              aws.ec2.vpc_endpoint.present:
                - vpc_id: 'string'
                - service_name: 'string'
                - resource_id: 'string'
                - vpc_endpoint_type: 'string'
                - policy_document: 'string'
                - route_table_ids:
                  - 'string'
                - subnet_ids:
                  - 'string'
                - security_group_ids:
                  - 'string'
                - private_dns_enabled: 'Boolean'
                - client_token: 'string'
                - tag_specifications:
                - Key: 'string'
                  Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            vpce-1231231234:
              aws.ec2.vpc_endpoint.present:
                - vpc_id: vpc-123123123
                - service_name: com.amazonaws.us-west-2.s3
                - resource_id: vpce-1231231234
                - vpc_endpoint_type: Interface
                - private_dns_enabled: False
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        before = await hub.exec.aws.ec2.vpc_endpoint.get(
            ctx, resource_id=resource_id, name=name
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        old_tags = result["old_state"].get("tags", None)
        plan_state = copy.deepcopy(result["old_state"])

        is_tag_update_required = tags is not None and tags != old_tags
        if is_tag_update_required:
            update_ret = await hub.tool.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=old_tags,
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            if not update_ret["result"]:
                result["result"] = False
                return result
            if ctx.get("test", False) and update_ret["result"]:
                plan_state["tags"] = update_ret["ret"]

        hub.log.debug(f"Updating existing aws.ec2.vpc_endpoint '{name}'")
        ret = await hub.tool.aws.ec2.vpc_endpoint.update_vpc_endpoint(
            ctx,
            result["old_state"],
            resource_id=resource_id,
            policy_document=policy_document,
            route_table_ids=route_table_ids,
            subnet_ids=subnet_ids,
            security_group_ids=security_group_ids,
            private_dns_enabled=private_dns_enabled,
            timeout=timeout,
        )
        result["result"] = ret["result"]
        result["comment"] = ret["comment"]

        if not result["result"]:
            return result
        # if no update required then ret["result"] is True but ret["ret"] is None
        if not is_tag_update_required and not ret["ret"]:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.ec2.vpc_endpoint", name=name
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result
        if ctx.get("test", False):
            result["new_state"] = (
                ret["ret"] if ret["ret"] else copy.deepcopy(result["old_state"])
            )
            if plan_state.get("tags") is not None:
                result["new_state"]["tags"] = plan_state["tags"]
            return result
    else:
        if ctx.get("test", False):
            plan_state = {
                "name": name,
                "vpc_id": vpc_id,
                "service_name": service_name,
                "vpc_endpoint_type": vpc_endpoint_type,
                "policy_document": policy_document,
                "route_table_ids": route_table_ids,
                "subnet_ids": subnet_ids,
                "security_group_ids": security_group_ids,
                "private_dns_enabled": private_dns_enabled,
                "client_token": client_token,
                "tags": tags,
                "resource_id": "resource_id_known_after_present",
            }
            result["new_state"] = {k: v for k, v in plan_state.items() if v is not None}
            result["comment"] = (
                f"Would create aws.ec2.vpc_endpoint '{name}' and attach to vpc '{vpc_id}'",
            )
            return result

        hub.log.debug(f"Creating new aws.ec2.vpc_endpoint '{name}'")
        ret = await hub.exec.boto3.client.ec2.create_vpc_endpoint(
            ctx,
            ClientToken=client_token,
            VpcEndpointType=vpc_endpoint_type,
            VpcId=vpc_id,
            ServiceName=service_name,
            PolicyDocument=policy_document,
            RouteTableIds=route_table_ids,
            SubnetIds=subnet_ids,
            SecurityGroupIds=security_group_ids,
            PrivateDnsEnabled=private_dns_enabled,
            TagSpecifications=[
                {
                    "ResourceType": "vpc-endpoint",
                    "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                }
            ]
            if tags
            else None,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["VpcEndpoint"]["VpcEndpointId"]
        hub.log.debug(
            f"Created new aws.ec2.vpc_endpoint '{name}' with resource_id: '{resource_id}'"
        )
        result["comment"] = (
            f"Created aws.ec2.vpc_endpoint: '{name}' attached to vpc: '{vpc_id}'",
        )
    try:
        after = await hub.exec.aws.ec2.vpc_endpoint.get(
            ctx, resource_id=resource_id, name=name
        )
        if not after["result"] or not after["ret"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, timeout: Dict = None
) -> Dict[str, Any]:
    """
    Deletes the specified VPC endpoint.

    You can delete any of the following types of VPC endpoints. Gateway endpoint,   Gateway Load Balancer endpoint,
    Interface endpoint.

    The following rules apply when you delete a VPC endpoint:
        * When you delete a gateway endpoint, we delete the endpoint routes in the route tables that are associated with the endpoint.
        * When you delete a Gateway Load Balancer endpoint, we delete the endpoint network interfaces. You can only delete
          Gateway Load Balancer endpoints when the routes that are associated with the endpoint are deleted.
        * When you delete an interface endpoint, we delete the endpoint network interfaces.

    Args:
        name(str):
            An idem name of the resource

        resource_id(str, Optional):
            An AWS vpc_endpoint resource_id to identify the resource.

        timeout(dict, Optional):
            Timeout configuration for deleting the endpoint.

            * delete (dict):
                Timeout configuration for deleting the endpoint.

                * delay:
                    The amount of time in seconds to wait between attempts.

                * max_attempts:
                    Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        .. code-block:: sls

            [vpc_endpoint-resource-id]:
              aws.ec2.vpc_endpoint.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            vpce-123123123:
              aws.ec2.vpc_endpoint.absent:
                - name: value
                - resource_id: vpce-123123123
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.ec2.vpc_endpoint", name=name
    )
    if not resource_id:
        result["comment"] = resource_already_absent_msg
        return result
    before = await hub.exec.aws.ec2.vpc_endpoint.get(
        ctx, resource_id=resource_id, name=name
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    hub.log.debug(f"Attempting to delete aws.ec2.vpc_endpoint '{name}'")

    if not before["ret"]:
        result["comment"] = resource_already_absent_msg
        return result
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = (f"Would delete aws.ec2.vpc_endpoint '{name}'",)
        return result
    else:
        if before["ret"]["state"] == "deleted":
            result["comment"] = resource_already_absent_msg
            return result
        ret = await hub.exec.boto3.client.ec2.delete_vpc_endpoints(
            ctx, VpcEndpointIds=[resource_id]
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + (ret["comment"],)
            return result
        result["comment"] = result["comment"] + (f"Deleted '{name}'",)

        delete_waiter_acceptors = [
            {
                "matcher": "pathAll",
                "expected": "available",
                "state": "retry",
                "argument": "VpcEndpoints[].State",
            },
            {
                "matcher": "pathAll",
                "expected": "pending",
                "state": "retry",
                "argument": "VpcEndpoints[].State",
            },
            {
                "matcher": "pathAll",
                "expected": "deleted",
                "state": "success",
                "argument": "VpcEndpoints[].State",
            },
            {
                "matcher": "error",
                "expected": "InvalidVpcEndpointId.NotFound",
                "state": "success",
                "argument": "Error.Code",
            },
        ]
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        endpoint_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="VpcEndpointDelete",
            operation="DescribeVpcEndpoints",
            argument=["Error.Code", "VpcEndpoints[].State"],
            acceptors=delete_waiter_acceptors,
            client=await hub.tool.boto3.client.get_client(ctx, "ec2"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "ec2",
                "VpcEndpointDelete",
                endpoint_waiter,
                VpcEndpointIds=[resource_id],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
    result["old_state"] = before["ret"]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Describes one or more of your VPC endpoints.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.vpc_endpoint
    """

    result = {}
    ret = await hub.exec.aws.ec2.vpc_endpoint.list(
        ctx, name="aws.ec2.vpc_endpoint.describe"
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.ec2.vpc_endpoint {ret['comment']}")
        return result

    for resource in ret["ret"]:
        result[resource.get("resource_id")] = {
            "aws.ec2.vpc_endpoint.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""Exec module for managing EC2 VPC Endpoint Service Permissions."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["soft_fail"]

__func_alias__ = {"list_": "list"}


async def get(
    hub, ctx, service_id: str, principal_arn: str, name: str = None
) -> Dict[str, Any]:
    """
    Describes the principal (service consumer) that are permitted to discover your VPC endpoint service.

    Args:
        service_id(str): The ID of the service.

        principal_arn(str): The ARN of the principal.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            unmanaged_resource:
              exec.run:
                - path: aws.ec2.vpc_endpoint_service_permission.get
                - kwargs:
                    - service_id: value
                    - principal_arn: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_permission.get service_id=value principal_arn=value
    """

    result = dict(comment=[], ret=None, result=True)

    get = await hub.exec.boto3.client.ec2.describe_vpc_endpoint_service_permissions(
        ctx,
        **{
            "ServiceId": service_id,
            # Create filter for principal_arn
            "Filters": [
                {"Name": "principal", "Values": [principal_arn]},
            ],
        },
    )

    # Case: Error
    if not get["result"]:
        if "NotFound" in str(get["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.vpc_endpoint_service_permission",
                    name=service_id,
                )
            )
            result["comment"].append(get["comment"])
            return result

        result["comment"].append(get["comment"])
        result["result"] = False
        return result

    # Case: Empty results
    if not get["ret"] or not get["ret"]["AllowedPrincipals"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.vpc_endpoint_service_permission",
                name=service_id,
            )
        )
        return result

    if len(get["ret"]["AllowedPrincipals"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_id=service_id,
                resource_type="aws.ec2.vpc_endpoint_service_permission",
            )
        )

    # Taking first one
    raw_resource = get["ret"]["AllowedPrincipals"][0]

    result[
        "ret"
    ] = await hub.tool.aws.ec2.vpc_endpoint_service_permission.convert_raw_resource_to_present_async(
        ctx=ctx,
        resource_id=None,
        raw_resource=raw_resource,
        idem_resource_name=name,
    )

    return result


async def list_(
    hub,
    ctx,
    service_id: str,
    filters: List[
        make_dataclass(
            "Filter",
            [
                ("Name", str, field(default=None)),
                ("Values", List[str], field(default=None)),
            ],
        )
    ] = None,
) -> Dict[str, Any]:
    """
    Describes the principals (service consumers) that are permitted to discover your VPC endpoint service.

    Args:
        service_id(str): The ID of the service.

        filters(List[dict[str, Any]], Optional): The filters.

            * principal - The ARN of the principal.

            * principal-type - The principal type (All | Service | OrganizationUnit | Account | User | Role). Defaults to None.

            (structure)
             A filter name and value pair that is used to return a more specific list of results from a describe operation. Filters can be used to match a set of resources by specific criteria, such as tags, attributes, or IDs.

             If you specify multiple filters, the filters are joined with an AND , and the request returns only results that match all of the specified filters.

                 * Name(str): The name of the filter. Filter names are case-sensitive.

                 * Values(List[str]): The filter values. Filter values are case-sensitive. If you specify multiple values for a filter, the values are joined with an OR , and the request returns all results that match any of the specified values.

    Returns:
        Dict[str, Any]

    Examples:

        Resource State:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.ec2.vpc_endpoint_service_permission.list
                - kwargs:
                    - service_id: value


        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_permission.list service_id=value
    """

    result = dict(comment=[], ret=[], result=True)

    ret = await hub.exec.boto3.client.ec2.describe_vpc_endpoint_service_permissions(
        ctx, **{"ServiceId": service_id, "Filters": filters}
    )

    if not ret["result"]:
        result["comment"].append(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"].get("AllowedPrincipals"):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.vpc_endpoint_service_permission", name=None
            )
        )
        return result

    for resource in ret["ret"]["AllowedPrincipals"]:
        result["ret"].append(
            await hub.tool.aws.ec2.vpc_endpoint_service_permission.convert_raw_resource_to_present_async(
                ctx=ctx,
                resource_id=None,
                raw_resource=resource,
                idem_resource_name=None,
            )
        )

    return result


async def create(
    hub, ctx, service_id: str, add_allowed_principals: List[str], name: str = None
) -> Dict[str, Any]:
    """
    Adds the permissions for your VPC endpoint service. You can add permissions for service consumers
    (Amazon Web Services accounts, users, and IAM roles) to connect to your endpoint service.

    If you grant permissions to all principals, the service is public. Any users who know the name of a public service can send a request to attach an endpoint. If the service does not require manual approval, attachments are automatically approved.

    Args:
        service_id(str): The ID of the service.

        add_allowed_principals(List[str]): The Amazon Resource Names (ARN) of the principals. Permissions are granted to the principals in this list. To grant permissions to all principals, specify an asterisk (*).

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.ec2.vpc_endpoint_service_permission.create:
                - kwargs:
                    - service_id: value
                    - add_allowed_principals:
                      - value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_permission.create service_id=value add_allowed_principals=[values]
    """

    result = dict(comment=[], ret={}, result=True)

    ret = await hub.exec.boto3.client.ec2.modify_vpc_endpoint_service_permissions(
        ctx,
        **{
            "ServiceId": service_id,
            "AddAllowedPrincipals": add_allowed_principals,
        },
    )

    result["result"] = ret["result"]

    if not result["result"]:
        result["comment"].append(ret["comment"])
        return result

    result["comment"].append(
        f"Created aws.ec2.vpc_endpoint_service_permission '{name}'",
    )

    result["ret"]["resource_id"] = ret.get("ret", {}).get("ServicePermissionId")
    result["ret"]["name"] = name

    return result


async def update(
    hub,
    ctx,
    service_id: str,
    add_allowed_principals: List[str] = None,
    remove_allowed_principals: List[str] = None,
    name: str = None,
) -> Dict[str, Any]:
    """
    Modifies the permissions for your VPC endpoint service. You can add or remove permissions for service consumers
    (Amazon Web Services accounts, users, and IAM roles) to connect to your endpoint service.

    If you grant permissions to all principals, the service is public. Any users who know the name of a public service can send a
    request to attach an endpoint. If the service does not require manual approval, attachments are automatically
    approved.

    Args:
        service_id(str): The ID of the service.

        add_allowed_principals(List[str], Optional): The Amazon Resource Names (ARN) of the principals. Permissions are granted to the principals in this list. To grant permissions to all principals, specify an asterisk (*). Defaults to None.

        remove_allowed_principals(List[str], Optional): The Amazon Resource Names (ARN) of the principals.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.ec2.vpc_endpoint_service_permission.update:
                - kwargs:
                    - service_id: value
                    - add_allowed_principals:
                        - value
                    - remove_allowed_principals:
                        - value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_permission.update service_id=value add_allowed_principals=[values] remove_allowed_principals=[values]
    """

    result = dict(comment=[], ret={}, result=True)

    # If at least one of them is given
    if add_allowed_principals or remove_allowed_principals:
        ret = await hub.exec.boto3.client.ec2.modify_vpc_endpoint_service_permissions(
            ctx,
            **{
                "ServiceId": service_id,
                "AddAllowedPrincipals": add_allowed_principals,
                "RemoveAllowedPrincipals": remove_allowed_principals,
            },
        )

        if not ret["result"]:
            result["result"] = False
            result["comment"].append(
                f"Could not update aws.ec2.vpc_endpoint_service_permission '{name}'",
            )
            result["comment"].append(ret["comment"])
            return result

        result["comment"].append(
            f"Updated aws.ec2.vpc_endpoint_service_permission '{name}'",
        )

    if "AddedPrincipals" in ret["ret"]:
        result["ret"]["resource_id"] = ret.get("ret", {}).get("ServicePermissionId")

    result["ret"]["name"] = name

    return result


async def delete(
    hub,
    ctx,
    service_id: str,
    remove_allowed_principals: List[str] = None,
    name: str = None,
) -> Dict[str, Any]:
    """
    Removes the permissions for your VPC endpoint service. You can remove permissions for service consumers
    (Amazon Web Services accounts, users, and IAM roles) to connect to your endpoint service.

    Args:
        service_id(str): The ID of the service.

        remove_allowed_principals(List[str], Optional): The Amazon Resource Names (ARN) of the principals.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.ec2.vpc_endpoint_service_permission.absent:
                - service_id: value
                - remove_allowed_principals:
                    - value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_permission.delete service_id=value remove_allowed_principals=[values]
    """

    result = dict(comment=[], ret=None, result=True)

    delete_ret = (
        await hub.exec.boto3.client.ec2.modify_vpc_endpoint_service_permissions(
            ctx,
            **{
                "ServiceId": service_id,
                "RemoveAllowedPrincipals": remove_allowed_principals,
            },
        )
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        result["comment"] = delete_ret["comment"]
        result["result"] = False
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.ec2.vpc_endpoint_service_permission", name=name
    )

    return result



"""Utility functions for Rds Db Proxy Target Groups."""
from typing import Any
from typing import Dict


async def convert_raw_resource_to_present_async(
    hub, ctx, idem_resource_name: str, resource_id: str, raw_resource: dict
) -> Dict[str, Any]:
    r"""
    Convert raw resource of db_proxy_target_group type into present format.

    """

    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}

    resource_parameters = {
        "ConnectionPoolConfig": "connection_pool_config",
        "DBProxyName": "db_proxy_name",
        "IsDefault": "is_default",
        "TargetGroupName": "target_group_name",
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



"""Exec functions for AWS API Gateway v2 API resources."""
from typing import Any
from typing import Dict


__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, api_id: str, route_id: str) -> Dict[str, Any]:
    """Get a route resource from AWS.

    Args:
        name(str):
            The name of the Idem state.

        api_id(str):
            The API identifier.

        route_id(str):
            The route ID.

    Returns:
        Dict[str, Any]:
            Returns a route.

    Examples:
        Call from the CLI:

        .. code-block:: bash

            idem exec aws.apigatewayv2.route.get name="my_resource" api_id="api_id" route_id="route_id"

        Call from code:

        .. code-block:: python

            async def my_func(hub, ctx, name:str, api_id:str, route_id:str):
                await hub.exec.aws.apigatewayv2.route.get(ctx, name, api_id, route_id)

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigatewayv2.route.get
                - kwargs:
                    name: my_resource
                    api_id: api_id
                    route_id: route_id
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.boto3.client.apigatewayv2.get_route(
        ctx, ApiId=api_id, RouteId=route_id
    )
    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigatewayv2.route", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result["ret"] = hub.tool.aws.apigatewayv2.route.convert_raw_route_to_present(
        raw_resource=ret["ret"],
        api_id=api_id,
    )
    return result


async def list_(hub, ctx, api_id: str) -> Dict[str, Any]:
    """Get a list of route resources from AWS.

    Args:
        api_id(str):
            The API identifier.

    Returns:
        Dict[str, Any]:
            Returns a route.

    Examples:
        Call from the CLI:

        .. code-block:: bash

            idem exec aws.apigatewayv2.route.get name="my_resource" api_id="api_id"

        Call from code:

        .. code-block:: python

            async def my_func(hub, ctx, name:str, api_id:str):
                await hub.exec.aws.apigatewayv2.route.get(ctx, name, api_id)

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigatewayv2.route.get
                - kwargs:
                    name: my_resource
                    api_id: api_id
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.boto3.client.apigatewayv2.get_routes(ctx, ApiId=api_id)

    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigatewayv2.route", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result["ret"] = []
    for item in ret["ret"]["Items"]:
        result["ret"].append(
            hub.tool.aws.apigatewayv2.route.convert_raw_route_to_present(
                raw_resource=item,
                api_id=api_id,
            )
        )
    return result



"""Exec module for managing Amazon Cloudfront Cache Policy."""

__func_alias__ = {"list_": "list"}

from typing import Dict


async def get(
    hub,
    ctx,
    name,
    resource_id: str = None,
) -> Dict:
    """Gets Cloudfront cache policy from AWS account.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, Optional):
            AWS cloudfront cache policy id to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.cloudfront.cache_policy.get name="idem_name" resource_id="resource_id"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.cloudfront.cache_policy.get
                - kwargs:
                    name: my_resource
                    resource_id: resource_id
    """

    result = dict(comment=[], ret=None, result=True)

    ret_list = await hub.exec.aws.cloudfront.cache_policy.list(ctx)

    if not ret_list["result"]:
        result["comment"] += list(ret_list["comment"])
        result["result"] = False
        hub.log.debug(f"Could not list cache policies {ret_list['comment']}")
        return result

    if not ret_list.get("ret"):
        result["comment"] += ("No cache policy present",)
        return result

    for policy in ret_list["ret"]:
        if (resource_id and policy.get("resource_id") == resource_id) or policy.get(
            "name"
        ) == name:
            result["ret"] = policy
            break

    if not result["ret"]:
        result["comment"] += (f"Cache policy '{name}' not present",)

    return result


async def list_(
    hub,
    ctx,
) -> Dict:
    """Lists all Cloudfront cache policies.

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.cloudfront.cache_policy.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.cloudfront.cache_policy.get

    """

    result = dict(comment=[], ret=[], result=True)

    ret_list = await hub.exec.boto3.client.cloudfront.list_cache_policies(ctx)

    if not ret_list["result"]:
        result["comment"] += list(ret_list["comment"])
        result["result"] = False
        hub.log.debug(f"Could not list cache policies {ret_list['comment']}")
        return result

    if not ret_list.get("ret"):
        result["comment"] += ("No cache policy present",)
        return result

    items = ret_list["ret"]["CachePolicyList"]["Items"]

    for item in items:
        policy_id = item["CachePolicy"]["Id"]
        name = item["CachePolicy"]["CachePolicyConfig"]["Name"]

        ret = await hub.exec.boto3.client.cloudfront.get_cache_policy(ctx, Id=policy_id)
        if not ret["result"]:
            if "NoSuchCachePolicy" in str(ret["comment"]):
                result["comment"].append(
                    hub.tool.aws.comment_utils.get_empty_comment(
                        resource_type="aws.cloudfront.cache_policy", name=name
                    )
                )

            result["comment"] += list(ret["comment"])
            result["result"] = False
            continue

        if not ret["ret"]["CachePolicy"]:
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.cloudfront.cache_policy", name=name
                )
            )
            continue

        ret["ret"]["CachePolicy"]["ETag"] = ret["ret"]["ETag"]

        result["ret"].append(
            hub.tool.aws.cloudfront.conversion_utils.convert_raw_cache_policy_to_present(
                ctx, raw_resource=ret["ret"]["CachePolicy"], idem_resource_name=name
            )
        )

    return result



"""Utility functions for EC2 VPC Endpoint Service Permissions."""
from typing import Any
from typing import Dict


async def convert_raw_resource_to_present_async(
    hub, ctx, idem_resource_name: str, resource_id: str, raw_resource: dict
) -> Dict[str, Any]:
    r"""
    Convert raw resource of vpc_endpoint_service_permission type into present format.

    Args:
        idem_resource_name(str): An Idem name of the resource.

        resource_id(str): An identifier of the resource in the provider.

        raw_resource(dict): The raw representation of the resource in the provider.

    Returns:
        Dict[str, Any]
    """
    resource_translated = {}

    resource_parameters = {
        "PrincipalType": "principal_type",
        "Principal": "principal_arn",
        "ServicePermissionId": "service_permission_id",
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    resource_translated["name"] = idem_resource_name or raw_resource.get(
        "ServicePermissionId"
    )
    resource_translated["resource_id"] = (
        raw_resource.get("ServicePermissionId") or resource_id or idem_resource_name
    )

    if "Tags" in resource_translated:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )

    return resource_translated



"""State module for managing Amazon Configuration Aggregator."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    account_aggregation_sources: List[
        make_dataclass(
            "AccountAggregationSource",
            [
                ("AccountIds", List[str]),
                ("AllAwsRegions", bool, field(default=None)),
                ("AwsRegions", List[str], field(default=None)),
            ],
        )
    ] = None,
    organization_aggregation_source: make_dataclass(
        "OrganizationAggregationSource",
        [
            ("RoleArn", str),
            ("AwsRegions", List[str], field(default=None)),
            ("AllAwsRegions", bool, field(default=None)),
        ],
    ) = None,
    tags: List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    """Creates an Amazon Configuration Aggregator.

    Creates and updates the configuration aggregator with the selected source accounts and regions.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider, a name of Configuration Aggregator. Defaults to None.

        account_aggregation_sources(List[Dict[str, Any]], Optional):
            A list of AccountAggregationSource object. Defaults to None.
                * AccountIds (List[str]): The 12-digit account ID of the account being aggregated.
                * AllAwsRegions (bool, optional): If true, aggregate existing Config regions and future regions.
                * AwsRegions (List[str], optional): The source regions being aggregated.

        organization_aggregation_source(Dict[str, Any], Optional):
            An OrganizationAggregationSource object. Defaults to None.
                * RoleArn (str): ARN of the IAM role.
                * AwsRegions (List[str], optional): The source regions being aggregated.
                * AllAwsRegions (bool, optional): If true, aggregate existing Config regions and future regions.

        tags(List[Dict[str, Any]], Optional): An array of tag object. Defaults to None.
            * Key (str, optional): One part of a key-value pair that make up a tag. A key is a general label that acts like a
            category for more specific tag values.
            * Value (str, optional): The optional part of a key-value pair that make up a tag. A value acts as a descriptor within a
            tag category (key).

    Request Syntax:
        aws.config.configuration_aggregator.present:
            - name: 'string'
            - resource_id: 'string'
            - account_aggregation_sources:
                - AccountIds:
                    - 'string'
                  AllAwsRegions: True|False
                  AwsRegions:
                    - 'string'
            organization_aggregation_source:
                - RoleArn: 'string'
                  AwsRegions:
                    - 'string'
                  AwsRegions: True|False

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.config.configuration_aggregator.present:
                - name: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if not resource_id:
        resource_id = name
    if resource_id:
        before = await hub.exec.aws.config.configuration_aggregator.get(
            ctx, name=resource_id
        )
    if before["ret"]:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "account_aggregation_sources": account_aggregation_sources,
                    "organization_aggregation_source": organization_aggregation_source,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.config.configuration_aggregator", name=name
            )
            return result
        resource_updated = (
            await hub.tool.aws.config.config_utils.is_configuration_aggregator_updated(
                before=before["ret"],
                account_aggregation_sources=account_aggregation_sources,
                organization_aggregation_source=organization_aggregation_source,
            )
        )
        if resource_updated:
            ret = await hub.exec.boto3.client.config.put_configuration_aggregator(
                ctx,
                ConfigurationAggregatorName=name,
                AccountAggregationSources=account_aggregation_sources,
                OrganizationAggregationSource=organization_aggregation_source,
                Tags=tags,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.config.configuration_aggregator", name=name
            )
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.config.configuration_aggregator", name=name
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
        return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "account_aggregation_sources": account_aggregation_sources,
                    "organization_aggregation_source": organization_aggregation_source,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.config.configuration_aggregator", name=resource_id
            )
            return result

        ret = await hub.exec.boto3.client.config.put_configuration_aggregator(
            ctx,
            ConfigurationAggregatorName=name,
            AccountAggregationSources=account_aggregation_sources,
            OrganizationAggregationSource=organization_aggregation_source,
            Tags=tags,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.config.configuration_aggregator", name=resource_id
        )

        if before or resource_updated:
            resource_id = ret["ret"]["ConfigurationAggregator"][
                "ConfigurationAggregatorName"
            ]
            after = await hub.exec.aws.config.configuration_aggregator.get(
                ctx, name=resource_id
            )
            result["new_state"] = after["ret"]
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes the specified configuration aggregator and the aggregated data associated with the aggregator.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider.

    Request Syntax:
        aws.config.configuration_aggregator.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.config.configuration_aggregator.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        resource_id = name
    before = await hub.exec.aws.config.configuration_aggregator.get(
        ctx, name=resource_id
    )
    if not before["ret"]:
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.configuration_aggregator", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]

        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.config.configuration_aggregator", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]

        ret = await hub.exec.boto3.client.config.delete_configuration_aggregator(
            ctx, ConfigurationAggregatorName=name
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.config.configuration_aggregator", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Examples:
        .. code-block:: bash

            $ idem describe aws.config.configuration_aggregator
    """
    result = {}
    ret = await hub.exec.aws.config.configuration_aggregator.list(ctx)

    if not ret["result"]:
        hub.log.warn(f"Could not describe configuration_aggregator {ret['comment']}")
        return {}

    for configuration_aggregator in ret["ret"]:
        result[configuration_aggregator["name"]] = {
            "aws.config.configuration_aggregator.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in configuration_aggregator.items()
            ]
        }
    return result



from collections import OrderedDict
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools import data
from dict_tools import differ


async def compare_inputs_and_update_lambda_function(
    hub,
    ctx,
    name: str,
    code: make_dataclass(
        """The code for the function.""" "FunctionCode",
        [
            ("ZipFile", bytes, field(default=None)),
            ("S3Bucket", str, field(default=None)),
            ("S3Key", str, field(default=None)),
            ("S3ObjectVersion", str, field(default=None)),
            ("ImageUri", str, field(default=None)),
        ],
    ),
    resource_id: str = None,
    plan_state: Dict[str, Any] = None,
    old_state: Dict[str, Any] = None,
    zip_file: Dict[str, Any] = None,
    s3_bucket: str = None,
    s3_key: str = None,
    s3_object_version: str = None,
    image_uri: str = None,
    publish: str = None,
    revision_id: str = None,
    architectures: List[str] = None,
    role: str = None,
    handler: str = None,
    description: str = None,
    function_timeout: int = None,
    memory_size: int = None,
    vpc_config: make_dataclass(
        """Configuration for network connectivity to Amazon Web Services resources in a VPC"""
        "VpcConfig",
        [
            ("SubnetIds", List[str], field(default=None)),
            ("SecurityGroupIds", List[str], field(default=None)),
        ],
    ) = None,
    environment: make_dataclass(
        """Environment variables that are accessible from function code"""
        "Environment",
        [("Variables", Dict[str, str], field(default=None))],
    ) = None,
    runtime: str = None,
    dead_letter_config: make_dataclass(
        """A dead letter queue configuration to specify the queue or topic where Lambda sends asynchronous events"""
        "DeadLetterConfig",
        [("TargetArn", str, field(default=None))],
    ) = None,
    kms_key_arn: str = None,
    tracing_config: make_dataclass(
        """Config to sample and trace a subset of incoming requests with X-Ray."""
        "TracingConfig",
        [("Mode", str, field(default=None))],
    ) = None,
    layers: List[str] = None,
    file_system_configs: List[
        make_dataclass(
            """Config for connection settings for an Amazon EFS file system."""
            "FileSystemConfig",
            [("Arn", str), ("LocalMountPath", str)],
        )
    ] = None,
    image_config: make_dataclass(
        """Container image configuration values that override the values in the container image Dockerfile."""
        "ImageConfig",
        [
            ("EntryPoint", List[str], field(default=None)),
            ("Command", List[str], field(default=None)),
            ("WorkingDirectory", str, field(default=None)),
        ],
    ) = None,
    timeout: Dict = None,
):
    """
    Updates a Lambda function's code. If code signing is enabled for the function, the code package must be signed by
    a trusted publisher. The function's code is locked when you publish a version. You can't modify the code of a
    published version, only the unpublished version.

    Modify the version-specific settings of a Lambda function. When you update a function, Lambda provisions an
    instance of the function and its supporting resources. If your function connects to a VPC, this process can take
    a minute. During this time, you can't modify the function, but you can still invoke it. Updates the configuration
    for asynchronous invocation for a function, version, or alias.

    Update function's code, configuration, function's event invoke config This function compares data with new params
    and invokes update actions with appropriate arguments.

    Args:
        plan_state: idem --test state for update on AWS Lambda.
        name: The name of the Lambda function.
        code: a dictionary detailing existing code config.
        resource_id: The name/ ARN/ partial ARN of the AWS Lambda function, version, or alias.
        runtime(str, Optional): The identifier of the function's runtime. Runtime is required if the deployment package is a ``.zip`` file archive.
        revision_id(str, Optional): Only update the function if the revision ID matches the ID that's specified.
            Use this option to avoid modifying a function that has changed since you last read it.
        architectures(list, Optional): The instruction set architecture that the function supports. Enter a string array
            with one of the valid values (arm64 or x86_64). The default value is x86_64.
        image_config(Dict[str, Any], Optional):
            Container image configuration values that override the values in the container image Dockerfile. Defaults to None.
                * EntryPoint (list[str], Optional): Specifies the entry point to their application, which is typically the location of the runtime executable.
                * Command (list[str], Optional): Specifies parameters that you want to pass in with ENTRYPOINT.
                * WorkingDirectory (str, Optional): Specifies the working directory.
        file_system_configs(list[Dict[str, Any]], Optional):
            Connection settings for an Amazon EFS file system. Defaults to None.
                * Arn (str): The Amazon Resource Name (ARN) of the Amazon EFS access point that provides access to the file system.
                * LocalMountPath (str): The path where the function can access the file system, starting with /mnt/.
        layers(list, Optional): A list of function layers to add to the function's execution environment. Specify each layer by its ARN, including the version.
        tags(Dict[str, str], Optional): A list of tags to apply to the function. Defaults to None.
        tracing_config(Dict[str, Any], Optional):
            Set Mode to Active to sample and trace a subset of incoming requests with X-Ray. Defaults to None.
                * Mode (str, Optional): The tracing mode.
        kms_key_arn(str, Optional): The ARN of the Amazon Web Services Key Management Service (KMS) key that's used to encrypt your function's environment variables. If it's not provided, Lambda uses a default service key.
        environment(Dict[str, Any], Optional):
            Environment variables that are accessible from function code during execution. Defaults to None.
                * Variables (Dict[str, str], Optional): Environment variable key-value pairs. For more information, see Using Lambda environment variables.
        dead_letter_config(Dict[str, Any], Optional):
            A dead letter queue configuration that specifies the queue or topic where Lambda sends asynchronous events when they fail processing. Defaults to None.
                * TargetArn (str, Optional): The Amazon Resource Name (ARN) of an Amazon SQS queue or Amazon SNS topic.
                * package_type (str, Optional): The type of deployment package. Set to Image for container image and set Zip for ZIP archive.
        vpc_config(Dict[str, Any], Optional):
            For network connectivity to Amazon Web Services resources in a VPC, specify a list of security groups and subnets in the VPC. When you connect a function to a VPC, it can only access resources and the internet through that VPC. For more information, see VPC Settings. Defaults to None.
                * SubnetIds (list[str], Optional): A list of VPC subnet IDs.
                * SecurityGroupIds (list[str], Optional): A list of VPC security groups IDs.
        publish(bool, Optional): Set to true to publish the first version of the function during creation.
        memory_size(int, Optional): The amount of memory available to the function at runtime. Increasing the function memory also increases its CPU allocation.
            The default value is 128 MB. The value can be any multiple of 1 MB.
        function_timeout(int, Optional): The amount of time (in seconds) that Lambda allows a function to run before stopping it. The default is 3 seconds. The maximum allowed value is 900 seconds.
        description(str, Optional): A description of the function.
        handler(str, Optional): The name of the method within your code that Lambda calls to execute your function.
            Handler is required if the deployment package is a ``.zip`` file archive. The format includes the file name.
            It can also include namespaces and other qualifiers, depending on the runtime.
        timeout(Dict, Optional): Timeout configuration for create/update/deletion of AWS Lambda Function.
            * update (str) -- Timeout configuration for updating AWS Lambda Function
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)
    params_to_modify = {}
    modify_params = OrderedDict(
        {
            "Architectures": "architectures",
            "RevisionId": "revision_id",
            "ImageUri": "image_uri",
        }
    )
    for parameter_raw, parameter_present in modify_params.items():
        if locals()[parameter_present] is not None:
            if isinstance(locals()[parameter_present], list):
                if not hub.tool.aws.state_comparison_utils.are_lists_identical(
                    locals()[parameter_present], old_state.get(parameter_present)
                ):
                    params_to_modify[parameter_raw] = locals()[parameter_present]
            elif isinstance(locals()[parameter_present], Dict):
                if data.recursive_diff(
                    locals()[parameter_present],
                    old_state.get(parameter_present),
                    ignore_order=True,
                ):
                    params_to_modify[parameter_raw] = locals()[parameter_present]
            else:
                if locals()[parameter_present] != old_state.get(parameter_present):
                    params_to_modify[parameter_raw] = locals()[parameter_present]

    # update for lambda function code
    if params_to_modify:
        if ctx.get("test", False):
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.lambda.function", name=name
            )
            for key, value in params_to_modify.items():
                plan_state[modify_params.get(key)] = value
        else:
            modify_ret = await hub.exec.boto3.client["lambda"].update_function_code(
                ctx=ctx,
                FunctionName=resource_id,
                ZipFile=zip_file,
                S3Bucket=s3_bucket,
                S3Key=s3_key,
                S3ObjectVersion=s3_object_version,
                Publish=publish,
                **params_to_modify,
            )
            result["result"] = result["result"] and modify_ret["result"]
            result["comment"] += hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.lambda.function", name=name
            )
            if not modify_ret["result"]:
                result["comment"] = result["comment"] + modify_ret["comment"]
                result["result"] = False
                return result

            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=1,
                default_max_attempts=20,
                timeout_config=timeout.get("update") if timeout else None,
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "lambda",
                    "function_updated",
                    FunctionName=resource_id,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False

    # update for lambda function configuration
    params_to_modify_for_config = {}
    modify_params_for_config = OrderedDict(
        {
            "Role": "role",
            "Handler": "handler",
            "Description": "description",
            "Timeout": "function_timeout",
            "MemorySize": "memory_size",
            "VpcConfig": "vpc_config",
            "Environment": "environment",
            "Runtime": "runtime",
            "DeadLetterConfig": "dead_letter_config",
            "KMSKeyArn": "kms_key_arn",
            "TracingConfig": "tracing_config",
            "RevisionId": "revision_id",
            "Layers": "layers",
            "FileSystemConfigs": "file_system_configs",
            "ImageConfig": "image_config",
        }
    )
    for (
        parameter_raw_for_config,
        parameter_present_for_config,
    ) in modify_params_for_config.items():
        # Add to modify list only if parameter is changed
        if locals()[parameter_present_for_config] is not None:
            if isinstance(locals()[parameter_present_for_config], list):
                if not hub.tool.aws.state_comparison_utils.are_lists_identical(
                    locals()[parameter_present_for_config],
                    old_state.get(parameter_present_for_config),
                ):
                    params_to_modify_for_config[parameter_raw_for_config] = locals()[
                        parameter_present_for_config
                    ]
            elif isinstance(locals()[parameter_present_for_config], Dict):
                if data.recursive_diff(
                    locals()[parameter_present_for_config],
                    old_state.get(parameter_present_for_config),
                    ignore_order=True,
                ):
                    params_to_modify_for_config[parameter_raw_for_config] = locals()[
                        parameter_present_for_config
                    ]
            else:
                if locals()[parameter_present_for_config] != old_state.get(
                    parameter_present_for_config
                ):
                    params_to_modify_for_config[parameter_raw_for_config] = locals()[
                        parameter_present_for_config
                    ]

    if params_to_modify_for_config:
        if ctx.get("test", False):
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.lambda.function", name=name
            )
            for key, value in params_to_modify_for_config.items():
                plan_state[modify_params_for_config.get(key)] = value
        else:
            modify_ret_for_config = await hub.exec.boto3.client[
                "lambda"
            ].update_function_configuration(
                ctx=ctx, FunctionName=resource_id, **params_to_modify_for_config
            )
            result["result"] = result["result"] and modify_ret_for_config["result"]
            result["comment"] += hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.lambda.function", name=name
            )
            if not modify_ret_for_config["result"]:
                result["comment"] = result["comment"] + modify_ret_for_config["comment"]
                result["result"] = False
                return result

            waiter_config_for_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=1,
                default_max_attempts=20,
                timeout_config=timeout.get("update") if timeout else None,
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "lambda",
                    "function_updated",
                    FunctionName=resource_id,
                    WaiterConfig=waiter_config_for_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
    return result



"""Exec module for managing EC2 spot instance requests."""
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    filters: List = None,
) -> Dict:
    """
    Get a spot instance request resource from AWS. Supply one of the inputs as the filter.

    Args:
        name (str):
            The name of the Idem state.

        resource_id (str, Optional):
            ID of the spot instance request.

    Returns:
        Dict[str, Any]:
            Returns spot instance request in present format

    Examples:
        Calling this exec module function from the cli with filters

        .. code-block:: bash

            idem exec aws.ec2.spot_instance_request.get name="my_resource"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.spot_instance_request.get
                - kwargs:
                    name: my_resource
    """
    result = dict(comment=[], ret=None, result=True)
    boto3_filters = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=[{"name": "spot-instance-request-id", "values": [resource_id]}]
    )
    ret = await hub.exec.boto3.client.ec2.describe_spot_instance_requests(
        ctx=ctx,
        Filters=boto3_filters,
    )
    if not ret.get("ret", {}).get("SpotInstanceRequests"):
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.spot_instance_request", name=name
            )
        )
        return result

    resource = ret["ret"]["SpotInstanceRequests"][0]
    if len(ret["ret"]["SpotInstanceRequests"]) > 1:
        result["comment"].append(
            f"More than one aws.ec2.spot_instance_request resource was found. Use resource {resource.get('SpotInstanceRequestId')}"
        )

    result["ret"] = hub.tool.aws.ec2.conversion_utils.convert_raw_sir_to_present(
        resource
    )
    return result


async def list_(hub, ctx, name: str = None, filters: List = None) -> Dict:
    """
    Get a list of spot instance request resources from AWS. Supply one of the inputs as the filter.

    Args:
        name (str, Optional):
            The name of the Idem state.

        filters (list[str, str], Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_spot_instance_requests

    Returns:
        Dict[str, Any]:
            Returns spot instance request list in present format

    Examples:
        Calling this exec module function from the cli with filters

        .. code-block:: bash

            idem exec aws.ec2.spot_instance_request.list name="my_resource" filters=[{'name': 'name', 'values': ['resource-name']}]

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.spot_instance_request.list
                - kwargs:
                    name: my_resource
                    filters:
                        - name: 'name'
                          values: ['resource-name']
    """
    result = dict(comment=[], ret=[], result=True)
    boto3_filters = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    ret = await hub.exec.boto3.client.ec2.describe_spot_instance_requests(
        ctx,
        Filters=boto3_filters,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["SpotInstanceRequests"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.spot_instance_requests", name=name
            )
        )
        return result
    for spot_instance_request in ret["ret"]["SpotInstanceRequests"]:
        result["ret"].append(
            hub.tool.aws.ec2.conversion_utils.convert_raw_sir_to_present(
                spot_instance_request=spot_instance_request
            )
        )
    return result



from typing import Any
from typing import Dict


async def is_organization_configuration_updated(
    hub, before: Dict[str, Any], auto_enable: False | True, data_sources: Dict = None
):
    if auto_enable != before.get("auto_enable"):
        return True
    if data_sources:
        if data_sources.get("S3Logs").get("AutoEnable") != before.get(
            "data_sources"
        ).get("S3Logs").get("AutoEnable"):
            return True
        if data_sources.get("Kubernetes").get("AuditLogs").get(
            "AutoEnable"
        ) != before.get("data_sources").get("Kubernetes").get("AuditLogs").get(
            "AutoEnable"
        ):
            return True
        if data_sources.get("MalwareProtection").get("ScanEc2InstanceWithFindings").get(
            "EbsVolumes"
        ).get("AutoEnable") != before.get("data_sources").get("MalwareProtection").get(
            "ScanEc2InstanceWithFindings"
        ).get(
            "EbsVolumes"
        ).get(
            "AutoEnable"
        ):
            return True

    return False



"""
Exec module for rds.db_subnet_group
"""
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__func_alias__ = {"list_": "list"}


async def get(hub, ctx, resource_id: str, name: str = None) -> Dict[str, Any]:
    """
    Returns a list of DBSubnetGroup descriptions. If a DBSubnetGroupName is specified, the list will contain only the
    descriptions of the specified DBSubnetGroup. For an overview of CIDR ranges, go to the Wikipedia Tutorial.

        Args:
            resource_id(str):
                An identifier of the resource in the provider. Defaults to None.

            name(str, Optional):
                An Idem name of the resource

        Examples:

            .. code-block:: bash

                idem exec aws_auto.rds.db_subnet_group.get.get resource_id=01235789abcdef

            .. code-block:: python

                async def my_func(hub, ctx):
                    ret = await hub.exec.aws_auto.rds.db_subnet_group.get.get(ctx, resource_id="012345789abcedf")

            .. code-block:: yaml

                aws_auto_rds_db_subnet_group_get_resource:
                  exec.run:
                    - path: aws_auto.rds.db_subnet_group.get
                    - resource_id: 0123456789abcdef

    """

    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    ret = await hub.exec.boto3.client.rds.describe_db_subnet_groups(
        ctx=ctx,
        DBSubnetGroupName=resource_id,
    )
    if not ret["result"]:
        if "DBSubnetGroupNotFoundFault" in str(ret.get("comment", "")):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.rds.db_subnet_group",
                    name=name if name else resource_id,
                )
            )
        else:
            ret["result"] = False
        result["comment"] += list(ret["comment"])
        return result

    if not ret["ret"]["DBSubnetGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_subnet_group",
                name=name if name else resource_id,
            )
        )
        return result

    resource = ret["ret"]["DBSubnetGroups"][0]
    if len(ret["ret"]["DBSubnetGroups"]) > 1:
        result["comment"].append(
            f"More than one aws.rds.db_subnet_group resource was found. Use resource {resource.get('DBSubnetGroupName')}"
        )

    result[
        "ret"
    ] = hub.tool.aws.rds.conversion_utils.convert_raw_db_subnet_group_to_present(
        resource=resource
    )

    return result


async def list_(
    hub,
    ctx,
    name: str = None,
    filters: List[
        make_dataclass("Filter", [("Name", str), ("Values", List[str])])
    ] = None,
) -> Dict[str, Any]:
    """
    Returns a list of DBSubnetGroup descriptions. If a DBSubnetGroupName is specified, the list will contain only the
    descriptions of the specified DBSubnetGroup. For an overview of CIDR ranges, go to the Wikipedia Tutorial.

        Args:
            name(str):
                An Idem name of the resource.

            filters(list[Dict[str, Any]], Optional):
                This parameter isn't currently supported. Defaults to None.

                * Name (str):
                    The name of the filter. Filter names are case-sensitive.

                * Values (list[str]):
                    One or more filter values. Filter values are case-sensitive.

        Examples:

            .. code-block:: bash

                idem exec aws_auto.rds.db_subnet_group.list

            .. code-block:: python

                async def my_func(hub, ctx):
                    ret = await hub.exec.aws_auto.rds.db_subnet_group.list(ctx)

            .. code-block:: yaml

                aws_auto_rds_db_subnet_group_list_resource:
                  exec.run:
                    - path: aws_auto.rds.db_subnet_group.list

    """

    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    ret = await hub.exec.boto3.client.rds.describe_db_subnet_groups(
        ctx=ctx,
        filters=filters,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        return result
    if not ret["ret"]["DBSubnetGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_subnet_group", name=name
            )
        )
        return result

    result["ret"] = []
    for resource in ret["ret"]["DBSubnetGroups"]:
        result["ret"].append(
            hub.tool.aws.rds.conversion_utils.convert_raw_db_subnet_group_to_present(
                resource=resource
            )
        )
    return result



"""State module for managing RDS DB Parameter Group."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_parameter_group_family: str,
    description: str = None,
    resource_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
    parameters: List[
        make_dataclass(
            "Parameter",
            [
                ("ParameterName", str),
                ("ParameterValue", str),
                ("ApplyMethod", str, field(default=None)),
            ],
        )
    ] = None,
) -> Dict[str, Any]:
    """Creates a new DB parameter group.

    A DB parameter group is initially created with the default parameters for the
    database engine used by the DB instance. To provide custom values for any of the parameters, you can specify
    parameters.  Once you've created a DB parameter group, you can associate it with your DB instance.

    When you associate a new DB parameter group with a running DB instance, you need to reboot the DB instance without
    failover for the new DB parameter group and associated settings to take effect.

    After you create a DB parameter group, you should wait at least 5 minutes before creating your first DB instance
    that uses that DB parameter group as the default parameter group. This allows Amazon RDS to fully complete the
    create action before the parameter group is used as the default for a new DB instance. This is especially important
    for parameters that are critical when creating the default database for a DB instance, such as the character set
    for the default database defined by the character_set_database parameter.

    You can use the Parameter Groups option of the Amazon RDS console to verify that your DB parameter group has been
    created or modified.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        db_parameter_group_family(str):
            The DB parameter group family name. A DB parameter group can be associated with one and only one
            DB parameter group family, and can be applied only to a DB instance running a database engine
            and engine version compatible with that DB parameter group family.

            To list all of the available parameter group families for a DB engine, use the following command:

            .. code-block:: bash

                aws rds describe-db-engine-versions --query "DBEngineVersions[].DBParameterGroupFamily" --engine <engine>

            For example, to list all of the available parameter group families for the MySQL DB engine, use the
            following command:

            .. code-block:: bash

                aws rds describe-db-engine-versions --query "DBEngineVersions[].DBParameterGroupFamily" --engine mysql

            The output contains duplicates.

            The following are the valid DB engine values:

              * aurora (for MySQL 5.6-compatible Aurora)
              * aurora-mysql (for MySQL 5.7-compatible and MySQL 8.0-compatible Aurora)
              * aurora-postgresql
              * mariadb
              * mysql
              * oracle-ee
              * oracle-ee-cdb
              * oracle-se2
              * oracle-se2-cdb
              * postgres
              * sqlserver-ee
              * sqlserver-se
              * sqlserver-ex
              * sqlserver-web

        description(str):
            The description for the DB parameter group.

        tags(Dict or List, Optional):
            Dict in the format of {tag-key: tag-value} or List of tags in the format of [{"Key": tag-key, "Value":
            tag-value}] to associate with the DB parameter group.

            Each tag consists of a key name and an associated value. Defaults to None.

            * Key (str, Optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of
              127 Unicode characters. May not begin with aws:.

            * Value(str, Optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a
              maximum of 256 Unicode characters.

        parameters(List, Optional):
            Parameters in the DB parameter group. Defaults to None.
            Parameter support the following:
            * ParameterName (str): Specifies the name of the parameter.
            * ParameterValue (str): Specifies the value of the parameter.
            * ApplyMethod (str, Optional): Indicates when to apply parameter updates.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.rds.db_parameter_group.present:
              - name: 'test'
              - resource_id: 'test'
              - db_parameter_group_family: 'aurora-mysql5.7'
              - description: 'test'
              - tags:
                 - Key: 'test'
                   Value: 'test1'
                 - Key: 'test2'
                   Value: 'test2'
              - parameters:
                  - ParameterName: 'aurora_disable_hash_join'
                    ParameterValue: '1'
                    ApplyMethod: 'immediate'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    plan_state = None

    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )
    if resource_id:
        before = await hub.exec.aws.rds.db_parameter_group.get(
            ctx=ctx, name=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.rds.db_parameter_group", name=name
        )
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])

        update_tags_ret = None
        # Update tags
        if tags is not None:
            update_tags_ret = await hub.exec.aws.rds.tag.update_rds_tags(
                ctx=ctx,
                resource_arn=result["old_state"].get("db_parameter_group_arn"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] += update_tags_ret["comment"]
            result["result"] = update_tags_ret["result"]
            resource_updated = update_tags_ret["result"]
            if ctx.get("test", False) and update_tags_ret["ret"] is not None:
                plan_state["tags"] = update_tags_ret["ret"]
        if parameters is not None:
            update_ret = await hub.tool.aws.rds.parameters.modify_db_parameter_group(
                ctx=ctx,
                resource_name=result["old_state"].get("resource_id"),
                old_parameters=result["old_state"].get("parameters"),
                new_parameters=parameters,
            )
            result["comment"] += update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if ctx.get("test", False) and update_ret["ret"] is not None:
                plan_state["parameters"] = update_ret["ret"].get("parameters")
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    "db_parameter_group_family": db_parameter_group_family,
                    "description": description,
                    "tags": tags,
                    "parameters": parameters,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.rds.db_parameter_group", name=name
            )
            return result
        ret = await hub.exec.boto3.client.rds.create_db_parameter_group(
            ctx,
            **{
                "DBParameterGroupName": name,
                "DBParameterGroupFamily": db_parameter_group_family,
                "Description": description,
                "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            },
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        resource_id = ret["ret"]["DBParameterGroup"]["DBParameterGroupName"]
        result["new_state"] = {"name": name, "resource_id": resource_id}

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.rds.db_parameter_group", name=name
        )

        # Modify the db group parameters
        if parameters is not None:
            ret_parameter = await hub.exec.boto3.client.rds.describe_db_parameters(
                ctx, DBParameterGroupName=resource_id
            )
            if not ret_parameter["result"]:
                result["comment"] += ret_parameter["comment"]
                result["result"] = ret_parameter["result"]
            else:
                update_ret = (
                    await hub.tool.aws.rds.parameters.modify_db_parameter_group(
                        ctx=ctx,
                        resource_name=resource_id,
                        old_parameters=ret_parameter["ret"]["Parameters"],
                        new_parameters=parameters,
                    )
                )

                result["comment"] += update_ret["comment"]
                result["result"] = result["result"] and update_ret["result"]
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.aws.rds.db_parameter_group.get(
                ctx=ctx, name=resource_id
            )
            if not after["result"] or not after["ret"]:
                result["result"] = False
                result["comment"] += after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] += (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes a specified DB parameter group.

    The DB parameter group to be deleted can't be associated with any DB
    instances.

    Args:
        resource_id(str):
            An identifier of the resource in the provider.

        name(str, Optional):
            The name of the DB parameter group.

            Constraints:

              * Must be the name of an existing DB parameter group.
              * You can't delete a default DB parameter group.
              * Can't be associated with any DB instances.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.rds.db_parameter_group.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_parameter_group", name=name
        )
        return result
    before = await hub.exec.aws.rds.db_parameter_group.get(ctx=ctx, name=resource_id)
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_parameter_group", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.rds.db_parameter_group", name=name
        )
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.rds.delete_db_parameter_group(
            ctx, DBParameterGroupName=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.rds.db_parameter_group", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Returns a list of DB Parameter Group descriptions.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.rds.db_parameter_group
    """
    result = {}
    ret = await hub.exec.aws.rds.db_parameter_group.list(
        ctx, name="aws.rds.db_parameter_group.describe"
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe db_parameter_group {ret['comment']}")
        return {}

    for resource in ret["ret"]:
        result[resource.get("resource_id")] = {
            "aws.rds.db_parameter_group.present": [resource]
        }
    return result
